"""Chain that runs an arbitrary python function."""
import logging
from typing import Callable, Dict, Optional

import httpx
from autogen import UserProxyAgent
from langchain.base_language import BaseLanguageModel

logger = logging.getLogger(__name__)


class AutoGenUserProxyAgent(UserProxyAgent):
    """A proxy agent for the user, that can execute code and provide feedback to the other agents.
    """

    def __init__(
        self,
        name: str,
        max_consecutive_auto_reply: Optional[int] = None,
        human_input_mode: Optional[str] = 'ALWAYS',  # hmean feedback input
        code_execution_flag: Optional[bool] = True,  # code execution
        function_map: Optional[Dict[str, Callable]] = None,  # function call
        llm_flag: Optional[bool] = False,  # llm call
        model_name: Optional[str] = 'gpt-4-0613',  # when llm_flag=True, need to set
        openai_api_key: Optional[str] = '',  # when llm_flag=True, need to set
        openai_api_base: Optional[str] = '',  # when llm_flag=True, need to set
        openai_proxy: Optional[str] = '',  # when llm_flag=True, need to set
        temperature: Optional[float] = 0,  # when llm_flag=True, need to set
        api_type: Optional[str] = None,  # when llm_flag=True, need to set
        api_version: Optional[str] = None,  # when llm_flag=True, need to set
        llm: Optional[BaseLanguageModel] = None,
        system_message: Optional[
            str] = '',  # agent system message, llm or group chat manage will use
        is_termination_msg: Optional[Callable[[Dict], bool]] = None,
        **kwargs,
    ):
        is_termination_msg = (is_termination_msg if is_termination_msg is not None else
                              (lambda x: x.get('content') == 'TERMINATE'))

        if code_execution_flag:
            code_execution_config = {
                'work_dir': '/tmp/autogen_coding',  # code save path
                'use_docker': False,
            }
        else:
            code_execution_config = False

        if llm_flag:
            config_list = [
                {
                    'model': model_name,
                    'api_key': openai_api_key,
                    'api_base': openai_api_base,
                    'api_type': api_type,
                    'api_version': api_version,
                },
            ]
            if openai_proxy:
                config_list[0]['http_client'] = httpx.Client(proxies=openai_proxy)
                config_list[0]['http_async_client'] = httpx.AsyncClient(proxies=openai_proxy)

            if llm:
                llm_config = llm
            else:
                llm_config = {
                    'seed': 42,  # change the seed for different trials
                    'temperature': temperature,
                    'config_list': config_list,
                    'request_timeout': 120,
                }
        else:
            llm_config = False

        super().__init__(name,
                         is_termination_msg=is_termination_msg,
                         max_consecutive_auto_reply=max_consecutive_auto_reply,
                         human_input_mode=human_input_mode,
                         function_map=function_map,
                         code_execution_config=code_execution_config,
                         llm_config=llm_config,
                         system_message=system_message)


class AutoGenUser(UserProxyAgent):
    """A proxy agent for the user, that can provide feedback to the other agents.
    """

    def __init__(
        self,
        name: str,
        max_consecutive_auto_reply: Optional[int] = 10,
        human_input_mode: Optional[str] = 'ALWAYS',  # hmean feedback input
        system_message: Optional[
            str] = '',  # agent system message, llm or group chat manage will use
        is_termination_msg: Optional[Callable[[Dict], bool]] = None,
        **kwargs,
    ):
        is_termination_msg = (is_termination_msg if is_termination_msg is not None else
                              (lambda x: x.get('content') == 'TERMINATE'))
        code_execution_config = False
        llm_config = False

        super().__init__(name,
                         is_termination_msg=is_termination_msg,
                         max_consecutive_auto_reply=max_consecutive_auto_reply,
                         human_input_mode=human_input_mode,
                         code_execution_config=code_execution_config,
                         llm_config=llm_config,
                         system_message=system_message)


class AutoGenCoder(UserProxyAgent):
    """A proxy agent for the coder, that can execute code to the other agents.
    """

    def __init__(
        self,
        name: str,
        function_map: Optional[Dict[str, Callable]] = None,  # function call
        system_message: Optional[
            str] = '',  # agent system message, llm or group chat manage will use
        is_termination_msg: Optional[Callable[[Dict], bool]] = None,
        **kwargs,
    ):
        is_termination_msg = (is_termination_msg if is_termination_msg is not None else
                              (lambda x: x.get('content') == 'TERMINATE'))
        code_execution_config = {
            'work_dir': '/tmp/autogen_coding',  # code save path
            'use_docker': False,
        }
        llm_config = False

        super().__init__(name,
                         is_termination_msg=is_termination_msg,
                         human_input_mode='NEVER',
                         function_map=function_map,
                         code_execution_config=code_execution_config,
                         llm_config=llm_config,
                         system_message=system_message)



import json

import requests

from .types import ChatInput, ChatOutput, Choice, Message, Usage
from .utils import get_ts


class ChatCompletion(object):

    def __init__(self, group_id, api_key, **kwargs):
        ep_url = 'https://api.minimax.chat/v1/text/chatcompletion'
        self.endpoint = f'{ep_url}?GroupId={group_id}'
        self.headers = {
            'Authorization': f'Bearer {api_key}',
            'Content-Type': 'application/json'
        }

    def parseChunkDelta(self, chunk):
        decoded_data = chunk.decode('utf-8')
        parsed_data = json.loads(decoded_data[6:])
        delta_content = parsed_data['choices'][0]['delta']
        return delta_content

    def __call__(self, inp: ChatInput, verbose=False):
        messages = inp.messages
        model = inp.model
        top_p = 0.95 if inp.top_p is None else inp.top_p
        temperature = 0.9 if inp.temperature is None else inp.temperature
        stream = False if inp.stream is None else inp.stream
        max_tokens = 1024 if inp.max_tokens is None else inp.max_tokens
        if abs(temperature) <= 1e-6:
            temperature = 1e-6

        chat_messages = messages
        system_prompt = ('MM智能助理是一款由MinMax自研的，没有调用其他产品接口的大型语言'
                         '模型。MiniMax是一家中国科技公司，一直致力于进行大模型相关的研究。\n----\n')

        if messages[0].role == 'system':
            system_prompt = messages[0].content
            chat_messages = messages[1:]

        new_messages = []
        for m in chat_messages:
            role = 'USER'
            if m.role == 'system' or m.role == 'assistant':
                role = 'BOT'

            new_messages.append({'sender_type': role, 'text': m.content})

        #  role_meta is given, prompt must is not empty
        system_info = {}
        if system_prompt:
            system_info = {
                'prompt': system_prompt,
                'role_meta': {
                    'user_name': '用户',
                    'bot_name': 'MM智能助理'
                }
            }

        payload = {
            'model': model,
            'stream': stream,
            'use_standard_sse': True,
            'messages': new_messages,
            'temperature': temperature,
            'top_p': top_p,
            'tokens_to_generate': max_tokens
        }
        payload.update(system_info)

        if verbose:
            print('payload', payload)

        response = requests.post(self.endpoint,
                                 headers=self.headers,
                                 json=payload)

        req_type = 'chat.completion'
        status_message = 'success'
        status_code = response.status_code
        created = get_ts()
        choices = []
        usage = Usage()
        if status_code == 200:
            try:
                info = json.loads(response.text)
                if info['base_resp']['status_code'] == 0:
                    created = info['created']
                    # reply = info['reply']
                    choices = []
                    for s in info['choices']:
                        index = s['index']
                        finish_reason = s['finish_reason']
                        msg = Message(role='assistant', content=s['text'])
                        cho = Choice(index=index,
                                     message=msg,
                                     finish_reason=finish_reason)
                        choices.append(cho)
                    total_tokens = info['usage']['total_tokens']
                    usage = Usage(total_tokens=total_tokens)
                else:
                    status_code = info['base_resp']['status_code']
                    status_message = info['base_resp']['status_msg']

            except Exception as e:
                status_code = 401
                status_message = str(e)
        else:
            status_code = 400
            status_message = 'requests error'

        if status_code != 200:
            raise Exception(status_message)

        return ChatOutput(status_code=status_code,
                          status_message=status_message,
                          model=model,
                          object=req_type,
                          created=created,
                          choices=choices,
                          usage=usage)



import os
import pandas as pd

excel_folder = '/home/public/rag_benchmark_v1.0/rag_qa_gen_filter_origin'
save_folder = '/home/public/rag_benchmark_v1.0/rag_qa_gen_filter'
excel_files = os.listdir(excel_folder)

total_correct = 0
total_num = 0
for excel_file in excel_files:
    # if excel_file != '96_MA_IFC300_cn_100909_4000069803_R04_qa_gen.xlsx':
    #     continue
    if not excel_file.endswith('.xlsx'):
        continue
    excel_path = os.path.join(excel_folder, excel_file)
    df = pd.read_excel(excel_path)
    qa_info = df.to_dict('records')
    correct_qa = []
    for qa in qa_info:
        if 'Unnamed: 1' in qa:
            qes_key = 'Unnamed: 1'
        elif '问题类型' in qa:
            qes_key = '问题类型'
        else:
            raise ValueError('key not found')
        
        if 'Unnamed: 4' in qa:
            ans_key = 'Unnamed: 4'
        elif '答案类型' in qa:
            ans_key = '答案类型'
        else:
            raise ValueError('key not found')

        valid_qa = dict()
        for key in qa:
            valid_qa[key.strip()] = qa[key]

        total_num += 1
        if isinstance(qa[qes_key], str) and qa[qes_key].strip() == '正确' and isinstance(qa[ans_key], str) and qa[ans_key].strip() == '正确':
            total_correct += 1
            correct_qa.append(valid_qa)
        else:
            print(qa[qes_key], qa[ans_key])
            print(excel_file)

    df = pd.DataFrame(correct_qa)
    df.to_excel(os.path.join(save_folder, excel_file), index=False)

print('total_correct:', total_correct)
print('total_num:', total_num)




from __future__ import annotations

import logging
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import requests
from langchain.embeddings.base import Embeddings
from langchain.utils import get_from_dict_or_env
from langchain_core.pydantic_v1 import BaseModel, Field, root_validator
from tenacity import (before_sleep_log, retry, retry_if_exception_type, stop_after_attempt,
                      wait_exponential)

logger = logging.getLogger(__name__)


def _create_retry_decorator(embeddings: HostEmbeddings) -> Callable[[Any], Any]:
    min_seconds = 4
    max_seconds = 10
    # Wait 2^x * 1 second between each retry starting with
    # 4 seconds, then up to 10 seconds, then 10 seconds afterwards
    return retry(
        reraise=True,
        stop=stop_after_attempt(embeddings.max_retries),
        wait=wait_exponential(multiplier=1, min=min_seconds, max=max_seconds),
        retry=(retry_if_exception_type(Exception)),
        before_sleep=before_sleep_log(logger, logging.WARNING),
    )


def embed_with_retry(embeddings: HostEmbeddings, **kwargs: Any) -> Any:
    """Use tenacity to retry the embedding call."""
    retry_decorator = _create_retry_decorator(embeddings)

    @retry_decorator
    def _embed_with_retry(**kwargs: Any) -> Any:
        return embeddings.embed(**kwargs)

    return _embed_with_retry(**kwargs)


class HostEmbeddings(BaseModel, Embeddings):
    """host embedding models.
    """

    client: Optional[Any]  #: :meta private:
    """Model name to use."""
    model: str = 'embedding-host'
    host_base_url: str = None

    deployment: Optional[str] = 'default'

    embedding_ctx_length: Optional[int] = 6144
    """The maximum number of tokens to embed at once."""
    """Maximum number of texts to embed in each batch"""
    max_retries: Optional[int] = 6
    """Maximum number of retries to make when generating."""
    request_timeout: Optional[Union[float, Tuple[float, float]]] = 200
    """Timeout in seconds for the OpenAPI request."""

    model_kwargs: Optional[Dict[str, Any]] = Field(default_factory=dict)
    """Holds any model parameters valid for `create` call not explicitly specified."""

    verbose: Optional[bool] = False

    url_ep: Optional[str] = None

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key and python package exists in environment."""
        values['host_base_url'] = get_from_dict_or_env(values, 'host_base_url', 'HostBaseUrl')
        model = values['model']
        try:
            url = values['host_base_url']
            values['url_ep'] = f'{url}/{model}/infer'
        except Exception:
            raise Exception(f'Failed to set url ep failed for model {model}')

        try:
            values['client'] = requests.post
        except AttributeError:
            raise ValueError('Try upgrading it with `pip install --upgrade requests`.')
        return values

    @property
    def _invocation_params(self) -> Dict:
        api_args = {
            'model': self.model,
            'request_timeout': self.request_timeout,
            **self.model_kwargs,
        }
        return api_args

    def embed(self, texts: List[str], **kwargs) -> List[List[float]]:
        emb_type = kwargs.get('type', 'raw')
        inp = {'texts': texts, 'model': self.model, 'type': emb_type}
        if self.verbose:
            print('payload', inp)

        max_text_to_split = 200
        outp = None

        start_index = 0
        len_text = len(texts)
        while start_index < len_text:
            inp_local = {
                'texts': texts[start_index:min(start_index + max_text_to_split, len_text)],
                'model': self.model,
                'type': emb_type
            }
            try:
                outp_single = self.client(url=self.url_ep,
                                          json=inp_local,
                                          timeout=self.request_timeout).json()
                if outp is None:
                    outp = outp_single
                else:
                    outp['embeddings'] += outp_single['embeddings']
            except requests.exceptions.Timeout:
                raise Exception(f'timeout in host embedding infer, url=[{self.url_ep}]')
            except Exception as e:
                raise Exception(f'exception in host embedding infer: [{e}]')

            if outp_single['status_code'] != 200:
                raise ValueError(f"API returned an error: {outp['status_message']}")
            start_index += max_text_to_split
        return outp['embeddings']

    def embed_documents(self,
                        texts: List[str],
                        chunk_size: Optional[int] = 0) -> List[List[float]]:
        if not texts:
            return []
        """Embed search docs."""
        texts = [text for text in texts if text]
        embeddings = embed_with_retry(self, texts=texts, type='doc')
        return embeddings

    def embed_query(self, text: str) -> List[float]:
        embeddings = embed_with_retry(self, texts=[text], type='query')
        return embeddings[0]


class ME5Embedding(HostEmbeddings):
    model: str = 'multi-e5'
    embedding_ctx_length: int = 512


class BGEZhEmbedding(HostEmbeddings):
    model: str = 'bge-zh'
    embedding_ctx_length: int = 512


class GTEEmbedding(HostEmbeddings):
    model: str = 'gte'
    embedding_ctx_length: int = 512


class JINAEmbedding(HostEmbeddings):
    model: str = 'jina'
    embedding_ctx_length: int = 512


class CustomHostEmbedding(HostEmbeddings):
    model: str = Field('custom-embedding', alias='model')
    embedding_ctx_length: int = 512

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key and python package exists in environment."""
        values['host_base_url'] = get_from_dict_or_env(values, 'host_base_url', 'HostBaseUrl')
        try:
            values['url_ep'] = values['host_base_url']
        except Exception:
            raise Exception('Failed to set url ep for custom host embedding')

        try:
            values['client'] = requests.post
        except AttributeError:
            raise ValueError('Try upgrading it with `pip install --upgrade requests`.')
        return values



"""Chain that runs an arbitrary python function."""
import logging
from typing import List, Optional

import httpx
from autogen import Agent, GroupChat, GroupChatManager
from langchain.base_language import BaseLanguageModel

from .user import AutoGenUser

logger = logging.getLogger(__name__)


class AutoGenGroupChatManager(GroupChatManager):
    """A chat manager agent that can manage a group chat of multiple agents.
    """

    def __init__(
        self,
        agents: List[Agent],
        max_round: int = 50,
        llm: Optional[BaseLanguageModel] = None,
        model_name: Optional[str] = 'gpt-4-0613',
        openai_api_key: Optional[str] = '',
        openai_api_base: Optional[str] = '',
        openai_proxy: Optional[str] = '',
        temperature: Optional[float] = 0,
        api_type: Optional[str] = None,  # when llm_flag=True, need to set
        api_version: Optional[str] = None,  # when llm_flag=True, need to set
        name: Optional[str] = 'chat_manager',
        system_message: Optional[str] = 'Group chat manager.',
        **kwargs,
    ):
        if not any(isinstance(agent, AutoGenUser) for agent in agents):
            raise Exception('chat_manager must contains AutoGenUser')

        groupchat = GroupChat(agents=agents, messages=[], max_round=max_round)

        config_list = [
            {
                'model': model_name,
                'api_key': openai_api_key,
                'api_base': openai_api_base,
                'api_type': api_type,
                'api_version': api_version,
            },
        ]
        if openai_proxy:
            config_list[0]['http_client'] = httpx.Client(proxies=openai_proxy)
            config_list[0]['http_async_client'] = httpx.AsyncClient(proxies=openai_proxy)

        if llm:
            llm_config = llm
        else:
            llm_config = {
                'seed': 42,  # change the seed for different trials
                'temperature': temperature,
                'config_list': config_list,
                'request_timeout': 120,
            }

        super().__init__(
            groupchat=groupchat,
            llm_config=llm_config,
            name=name,
            system_message=system_message,
        )



import json

import requests

from .types import ChatInput, ChatOutput, Choice, Message, Usage
from .utils import get_ts


def get_access_token(api_key, sec_key):
    url = (f'https://aip.baidubce.com/oauth/2.0/token?'
           f'grant_type=client_credentials'
           f'&client_id={api_key}&client_secret={sec_key}')

    payload = json.dumps('')
    headers = {
        'Content-Type': 'application/json',
        'Accept': 'application/json'
    }

    response = requests.request('POST', url, headers=headers, data=payload)
    return response.json().get('access_token')


class ChatCompletion(object):

    def __init__(self, api_key, sec_key, **kwargs):
        self.api_key = api_key
        self.sec_key = sec_key
        self.ep_url = ('https://aip.baidubce.com/rpc/2.0/ai_custom/v1/'
                       'wenxinworkshop/chat/completions')
        self.ep_url_pro = ('https://aip.baidubce.com/rpc/2.0/ai_custom/v1/'
                           'wenxinworkshop/chat/completions_pro')

        self.ep_url_turbo = ('https://aip.baidubce.com/rpc/2.0/ai_custom/v1/'
                             'wenxinworkshop/chat/eb-instant')

        # token = get_access_token(api_key, sec_key)
        # self.endpoint = f"{self.ep_url}?access_token={token}"
        self.headers = {'Content-Type': 'application/json'}

    def __call__(self, inp: ChatInput, verbose=False):
        messages = inp.messages
        model = inp.model
        top_p = 0.8 if inp.top_p is None else inp.top_p
        temperature = 0.95 if inp.temperature is None else inp.temperature
        stream = False if inp.stream is None else inp.stream
        # max_tokens = 1024 if inp.max_tokens is None else inp.max_tokens

        system_content = ''
        new_messages = []
        for m in messages:
            role = m.role
            if role == 'system':
                system_content = m.content
                continue
            new_messages.append({'role': role, 'content': m.content})

        if system_content:
            new_messages[-1]['content'] = system_content + '\n' + new_messages[
                -1]['content']

        payload = {
            'stream': stream,
            'messages': new_messages,
            'temperature': temperature,
            'top_p': top_p
        }

        if verbose:
            print('payload', payload)

        token = get_access_token(self.api_key, self.sec_key)
        endpoint = f'{self.ep_url}?access_token={token}'
        if model == 'ernie-bot-turbo':
            endpoint = f'{self.ep_url_turbo}?access_token={token}'
        elif model == 'ernie-bot-4':
            endpoint = f'{self.ep_url_pro}?access_token={token}'

        response = requests.post(endpoint, headers=self.headers, json=payload)

        req_type = 'chat.completion'
        status_message = 'success'
        status_code = response.status_code
        created = get_ts()
        choices = []
        usage = Usage()
        if status_code == 200:
            try:
                info = json.loads(response.text)
                status_code = info.get('error_code', 200)
                status_message = info.get('error_msg', status_message)
                if status_code == 200:
                    created = info['created']
                    result = info['result']
                    finish_reason = 'default'
                    msg = Message(role='assistant', content=result)
                    choices = [
                        Choice(index=0,
                               message=msg,
                               finish_reason=finish_reason)
                    ]
                    usage = Usage(**info['usage'])
            except Exception as e:
                status_code = 401
                status_message = str(e)
        else:
            status_code = 400
            status_message = 'requests error'

        if status_code != 200:
            raise Exception(status_message)

        return ChatOutput(status_code=status_code,
                          status_message=status_message,
                          model=model,
                          object=req_type,
                          created=created,
                          choices=choices,
                          usage=usage)



from typing import List, Mapping

from langchain.chains.router.base import Chain, MultiRouteChain, RouterChain


class MultiRuleChain(MultiRouteChain):
    router_chain: RouterChain
    destination_chains: Mapping[str, Chain]
    default_chain: Chain
    output_variables: List[str]

    @property
    def output_keys(self) -> List[str]:
        return self.output_variables



from .host_embedding import (BGEZhEmbedding, CustomHostEmbedding, GTEEmbedding, HostEmbeddings,
                             ME5Embedding, JINAEmbedding)
from .wenxin import WenxinEmbeddings
from .huggingfacemultilingual import HuggingFaceMultilingualEmbeddings
from .huggingfacegte import HuggingFaceGteEmbeddings

__all__ = [
    'WenxinEmbeddings', 'ME5Embedding', 'BGEZhEmbedding', 'GTEEmbedding',
    'HostEmbeddings', 'CustomHostEmbedding', 'JINAEmbedding',
    'HuggingFaceMultilingualEmbeddings', 'HuggingFaceGteEmbeddings'
]



import os
import pdb
import pandas as pd
import requests
import json
from config import FAST_GPT_API_KEY, FAST_GPT_APP_KEY, FAST_GPT_TOKEN
from loguru import logger
from tqdm import tqdm

datasets = set()
dataset = {}

def add_dataset(name, dataset_id):
    """    根据dataset 创建set， 映射name到id    """
    datasets.add(name)
    dataset[name] = dataset_id
    return dataset_id

def query_dataset_from_fastgpt():
    """
    知识库列表
    curl --location --request GET 'http://localhost:3000/api/core/dataset/list?parentId=' \
    --header 'Authorization: Bearer {{authorization}}' \
    """
    url = 'https://cloud.fastgpt.in/api/core/dataset/list?parentId='
    headers = {
        'Authorization': f'Bearer {FAST_GPT_API_KEY}', 
    }
    logger.debug(f'Starting to query dataset from fastgpt')
    response = requests.get(url, headers=headers).json()
    print(response)
    datasets = response['data']
        
    for dataset in datasets:
        name = dataset['name']
        id = dataset['_id']
        add_dataset(name, id)
    return datasets

def create_dataset(name):
    """curl --location --request POST 'http://localhost:3000/api/core/dataset/create' \
    --header 'Authorization: Bearer {{authorization}}' \
    --header 'Content-Type: application/json' \
    --data-raw '{
        "parentId": null,
        "type": "dataset",
        "name":"测试",
        "intro":"介绍",
        "avatar": "",
        "vectorModel": "text-embedding-ada-002",
        "agentModel": "gpt-3.5-turbo-16k"
    }'
    """
    url = 'https://cloud.fastgpt.in/api/core/dataset/create'
    headers = {
        'Authorization': f'Bearer {FAST_GPT_API_KEY}',
        'Content-Type': 'application/json',
    }
    data = {
        'name': name,
        'vectorModel': 'text-embedding-ada-002',
        'agentModel': 'gpt-4-1106-preview',
    }
    logger.debug(f'Starting to create dataset {name}')
    response = requests.post(url, headers=headers, data=json.dumps(data)).json()
    print(response)
    return response

def query_from_dataset(dataset_id):
    """curl --location --request GET 'http://localhost:3000/api/core/dataset/detail?id=6593e137231a2be9c5603ba7' \
    --header 'Authorization: Bearer {{authorization}}' \
    """
    url = f'https://cloud.fastgpt.in/api/core/dataset/detail?id={dataset_id}'
    headers = {
        'Authorization': f'Bearer {FAST_GPT_API_KEY}',
    }
    response = requests.get(url, headers=headers).json()
    print(response)

def query_file_from_dataset(dataset_id):
    """curl --location --request POST 'http://localhost:3000/api/core/dataset/collection/list' \
    --header 'Authorization: Bearer {{authorization}}' \
    --header 'Content-Type: application/json' \
    --data-raw '{
        "pageNum":1,
        "pageSize": 10,
        "datasetId":"6593e137231a2be9c5603ba7",
        "parentId": null,
        "searchText":""
    }'
    """
    url = f'https://cloud.fastgpt.in/api/core/dataset/collection/list'
    headers = {
        'Authorization': f'Bearer {FAST_GPT_API_KEY}',
         'Content-Type': 'application/json',
    }
    data = {
        'datasetId': dataset_id,
    }
    response = requests.post(url, headers=headers, data=json.dumps(data)).json()['data']
    print(response)
    
    files = response['data']
    return files
    
def create_by_file(file_path, dataset_id):
    """
    curl --location --request POST 'http://localhost:3000/api/proApi/core/dataset/collection/create/file' \
    --header 'Authorization: Bearer {{authorization}}' \
    --form 'file=@"C:\\Users\\user\\Desktop\\fastgpt测试文件\\index.html"' \
    --form 'data="{\"datasetId\":\"6593e137231a2be9c5603ba7\",\"parentId\":null,\"trainingType\":\"chunk\",\"chunkSize\":512,\"chunkSplitter\":\"\",\"qaPrompt\":\"\",\"metadata\":{}}"'
    """
    url = 'https://cloud.fastgpt.in/api/proApi/core/dataset/collection/create/file'
    headers = {
        'Authorization': f'Bearer {FAST_GPT_API_KEY}',
    }
    data = {
        'datasetId': dataset_id,
        'parentId': None,
        'trainingType': 'chunk',
        'chunkSize': 512,
        'chunkSplitter': '',
        'qaPrompt': '',
        'metadata': {},
    }
    files = {
        'file': open(file_path, 'rb'),
    }
    logger.debug(f'Starting to create file {file_path}')
    response = requests.post(url, headers=headers, data={'data': json.dumps(data)}, files=files).json()
    print(response)
    return response

def upload_file_to_fastgpt(data_dir, excel_file, save_excel_file):
    df = pd.read_excel(excel_file)
    all_questions_info = list()
    # 遍历每一行
    for index, row in df.iterrows():
        # 遍历每一列
        question_info = dict()
        for column in df.columns:
            value = row[column]
            question_info[column] = value
            # print(f"Row {index}, Column {column} has value {value}")
        all_questions_info.append(question_info)
    
    for questions_info in tqdm(all_questions_info):
        collection_name = questions_info['知识库名']
        file_name = questions_info['文件名']
        file_type = questions_info['文件类型']
        
  
        # 没有知识库先创建
        if collection_name not in datasets:
            res = create_dataset(collection_name)
            dataset_id = res['data']
            add_dataset(collection_name, dataset_id)
            
        else:
            dataset_id = dataset[collection_name]
        
        #  查询知识库中是否有该文件
        files = query_file_from_dataset(dataset_id)
        has_file = False

        for file in files:
            # TODO: fix pdf name 乱码
            if 'pdf' in file['name']:
                has_file = True
                file_id = file['_id']
                break
        if not has_file:
            try:
                res = create_by_file(os.path.join(data_dir, file_name), dataset_id)
                file_id = res['data']['collectionId']
            except Exception as e:
                file_id = ''
                logger.warning(f'error in create file {file_name}: {e}')
                
        questions_info['dataset_id'] = dataset_id           
        
        # save excel
        df = pd.DataFrame(all_questions_info)
        df.to_excel(save_excel_file, index=False)

def re_config_dataset(dataset_id, app_id='65b0b043c8685c9a14443f29'):
    """抓包的应用config url"""
    url = 'https://cloud.fastgpt.in/api/core/app/form2Modules/fastgpt-universal'
    headers = {
        'authority': 'cloud.fastgpt.in',
        'accept': 'application/json, text/plain, */*',
        'accept-language': 'zh-CN,zh;q=0.9',
        'content-type': 'application/json',
        'cookie': FAST_GPT_TOKEN,
        'origin': 'https://cloud.fastgpt.in',
        'referer': f'https://cloud.fastgpt.in/app/detail?appId={app_id}',
        'sec-ch-ua': '"Not_A Brand";v="8", "Chromium";v="120", "Google Chrome";v="120"',
        'sec-ch-ua-mobile': '?0',
        'sec-ch-ua-platform': '"Windows"',
        'sec-fetch-dest': 'empty',
        'sec-fetch-mode': 'cors',
        'sec-fetch-site': 'same-origin',
        'token': FAST_GPT_TOKEN,
        'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36',
    }
    false = False
    true = True

    logger.debug(f'Starting to re_config_dataset {dataset_id}')

    data = json.dumps(
        {"modules":[{"name":"core.module.template.User guide","flowType":"userGuide","inputs":[{"key":"welcomeText","type":"hidden","label":"core.app.Welcome Text","value":""},{"key":"variables","type":"hidden","label":"core.app.Chat Variable","value":[]},{"key":"questionGuide","type":"hidden","label":"core.app.Question Guide","value":false},{"key":"tts","type":"hidden","label":"core.app.TTS","value":{"type":"web"}}],"outputs":[],"position":{"x":447.98520778293346,"y":721.4016845336229},"moduleId":"userGuide"},{"moduleId":"userChatInput","name":"core.module.template.Chat entrance","avatar":"/imgs/module/userChatInput.png","flowType":"questionInput","position":{"x":324.81436595478294,"y":1527.0012457753612},"inputs":[{"key":"userChatInput","type":"systemInput","valueType":"string","label":"core.module.input.label.user question","showTargetInApp":false,"showTargetInPlugin":false,"connected":false}],"outputs":[{"key":"userChatInput","label":"core.module.input.label.user question","type":"source","valueType":"string","targets":[{"moduleId":"vuc92c","key":"userChatInput"}]}]},{"moduleId":"datasetSearch","name":"core.module.template.Dataset search","avatar":"/imgs/module/db.png","flowType":"datasetSearchNode","showStatus":true,"position":{"x":1351.5043753345153,"y":947.0780385418003},"inputs":[{"key":"switch","type":"target","label":"core.module.input.label.switch","valueType":"any","showTargetInApp":true,"showTargetInPlugin":true,"connected":false},{"key":"datasets","type":"selectDataset","label":"关联的知识库","value":[{"datasetId":dataset_id,"vectorModel":{"model":"text-embedding-ada-002","name":"Embedding-2","defaultToken":512,"maxToken":3000,"weight":100,"inputPrice":0.002,"outputPrice":0}}],"valueType":"selectDataset","list":[],"required":true,"showTargetInApp":false,"showTargetInPlugin":true,"connected":false},{"key":"similarity","type":"hidden","label":"最低相关性","value":0.5,"valueType":"number","min":0,"max":1,"step":0.01,"markList":[{"label":"0","value":0},{"label":"1","value":1}],"showTargetInApp":false,"showTargetInPlugin":false,"connected":false},{"key":"limit","type":"hidden","label":"引用上限","description":"单次搜索最大的 Tokens 数量，中文约1字=1.7Tokens，英文约1字=1Tokens","value":8000,"valueType":"number","showTargetInApp":false,"showTargetInPlugin":false,"connected":false},{"key":"searchMode","type":"hidden","label":"core.dataset.search.Mode","valueType":"string","showTargetInApp":false,"showTargetInPlugin":false,"value":"mixedRecall","connected":false},{"key":"usingReRank","type":"hidden","label":"","valueType":"string","showTargetInApp":false,"showTargetInPlugin":false,"value":false,"connected":false},{"key":"datasetParamsModal","type":"selectDatasetParamsModal","label":"","valueType":"any","showTargetInApp":false,"showTargetInPlugin":false,"connected":false},{"key":"userChatInput","type":"target","label":"core.module.input.label.user question","required":true,"valueType":"string","showTargetInApp":true,"showTargetInPlugin":true,"connected":true}],"outputs":[{"key":"isEmpty","label":"搜索结果为空","type":"source","valueType":"boolean","targets":[]},{"key":"unEmpty","label":"搜索结果不为空","type":"source","valueType":"boolean","targets":[]},{"key":"quoteQA","label":"引用内容","description":"始终返回数组，如果希望搜索结果为空时执行额外操作，需要用到上面的两个输入以及目标模块的触发器","type":"source","valueType":"datasetQuote","targets":[{"moduleId":"chatModule","key":"quoteQA"}]},{"key":"finish","label":"core.module.output.label.running done","description":"core.module.output.description.running done","valueType":"boolean","type":"source","targets":[]},{"key":"userChatInput","label":"core.module.input.label.user question","type":"hidden","valueType":"string","targets":[{"moduleId":"chatModule","key":"userChatInput"}]}]},{"moduleId":"chatModule","name":"AI 对话","avatar":"/imgs/module/AI.png","flowType":"chatNode","showStatus":true,"position":{"x":2022.7264786978908,"y":1006.3102431257475},"inputs":[{"key":"switch","type":"target","label":"core.module.input.label.switch","valueType":"any","showTargetInApp":true,"showTargetInPlugin":true,"connected":false},{"key":"model","type":"selectChatModel","label":"core.module.input.label.aiModel","required":true,"valueType":"string","showTargetInApp":false,"showTargetInPlugin":false,"value":"gpt-4-1106-preview","connected":false},{"key":"temperature","type":"hidden","label":"温度","value":0,"valueType":"number","min":0,"max":10,"step":1,"markList":[{"label":"严谨","value":0},{"label":"发散","value":10}],"showTargetInApp":false,"showTargetInPlugin":false,"connected":false},{"key":"maxToken","type":"hidden","label":"回复上限","value":2000,"valueType":"number","min":100,"max":4000,"step":50,"markList":[{"label":"100","value":100},{"label":"4000","value":4000}],"showTargetInApp":false,"showTargetInPlugin":false,"connected":false},{"key":"isResponseAnswerText","type":"hidden","label":"返回AI内容","value":true,"valueType":"boolean","showTargetInApp":false,"showTargetInPlugin":false,"connected":false},{"key":"quoteTemplate","type":"hidden","label":"引用内容模板","valueType":"string","showTargetInApp":false,"showTargetInPlugin":false,"connected":false},{"key":"quotePrompt","type":"hidden","label":"引用内容提示词","valueType":"string","showTargetInApp":false,"showTargetInPlugin":false,"connected":false},{"key":"aiSettings","type":"aiSettings","label":"","valueType":"any","showTargetInApp":false,"showTargetInPlugin":false,"connected":false},{"key":"systemPrompt","type":"textarea","label":"core.ai.Prompt","max":300,"valueType":"string","description":"模型固定的引导词，通过调整该内容，可以引导模型聊天方向。该内容会被固定在上下文的开头。可使用变量，例如 {{language}}","placeholder":"模型固定的引导词，通过调整该内容，可以引导模型聊天方向。该内容会被固定在上下文的开头。可使用变量，例如 {{language}}","showTargetInApp":true,"showTargetInPlugin":true,"connected":false},{"key":"history","type":"numberInput","label":"core.module.input.label.chat history","required":true,"min":0,"max":30,"valueType":"chatHistory","value":6,"showTargetInApp":true,"showTargetInPlugin":true,"connected":false},{"key":"quoteQA","type":"target","label":"引用内容","description":"对象数组格式，结构：\n [{q:'问题',a:'回答'}]","valueType":"datasetQuote","showTargetInApp":true,"showTargetInPlugin":true,"connected":true},{"key":"userChatInput","type":"target","label":"core.module.input.label.user question","required":true,"valueType":"string","showTargetInApp":true,"showTargetInPlugin":true,"connected":true}],"outputs":[{"key":"answerText","label":"AI回复","description":"将在 stream 回复完毕后触发","valueType":"string","type":"source","targets":[]},{"key":"finish","label":"core.module.output.label.running done","description":"core.module.output.description.running done","valueType":"boolean","type":"source","targets":[]},{"key":"history","label":"新的上下文","description":"将本次回复内容拼接上历史记录，作为新的上下文返回","valueType":"chatHistory","type":"source","targets":[]}]},{"moduleId":"vuc92c","name":"core.module.template.cfr","avatar":"/imgs/module/cfr.svg","flowType":"cfr","showStatus":true,"position":{"x":758.2985382279098,"y":1124.6527309337314},"inputs":[{"key":"switch","type":"target","label":"core.module.input.label.switch","valueType":"any","showTargetInApp":true,"showTargetInPlugin":true,"connected":false},{"key":"model","type":"selectExtractModel","label":"core.module.input.label.aiModel","required":true,"valueType":"string","value":"gpt-3.5-turbo","showTargetInApp":false,"showTargetInPlugin":false,"connected":false},{"key":"systemPrompt","type":"textarea","label":"core.module.input.label.cfr background","max":300,"value":"","valueType":"string","description":"core.module.input.description.cfr background","placeholder":"core.module.input.placeholder.cfr background","showTargetInApp":true,"showTargetInPlugin":true,"connected":false},{"key":"history","type":"numberInput","label":"core.module.input.label.chat history","required":true,"min":0,"max":30,"valueType":"chatHistory","value":6,"showTargetInApp":true,"showTargetInPlugin":true,"connected":false},{"key":"userChatInput","type":"target","label":"core.module.input.label.user question","required":true,"valueType":"string","showTargetInApp":true,"showTargetInPlugin":true,"connected":true}],"outputs":[{"key":"system_text","label":"core.module.output.label.cfr result","valueType":"string","type":"source","targets":[{"moduleId":"datasetSearch","key":"userChatInput"}]}]}],"type":"simple","simpleTemplateId":"fastgpt-universal"})
    
    response = requests.post(url, headers=headers, data=data)
    logger.debug(response.json())

def get_answer_from_fastgpt(query, app_id='65b0b043c8685c9a14443f29'):
    """curl --location --request POST 'https://api.fastgpt.in/api/v1/chat/completions' \
    --header 'Authorization: Bearer fastgpt-xxxxxx' \
    --header 'Content-Type: application/json' \
    --data-raw '{
        "chatId": "abcd",
        "stream": false,
        "detail": false,
        "variables": {
            "uid": "asdfadsfasfd2323",
            "name": "张三"
        },
        "messages": [
            {
                "content": "导演是谁",
                "role": "user"
            }
        ]
    }'
    """
    url = 'https://api.fastgpt.in/api/v1/chat/completions'
    headers = {
        'Authorization': f'Bearer {FAST_GPT_APP_KEY}',
        'Content-Type': 'application/json',
    }
    data = {
        "chatId": app_id,
        "stream": False,
        "detail": False,
        "variables": {
            "uid": "asdfadsfasfd2323",
            "name": "张三"
        },
        "messages": [
            {
                "content": query,
                "role": "user"
            }
        ]
    }
    response = requests.post(url, headers=headers, data=json.dumps(data)).json()
    print(response)
    return response

def search_dataset_from_fastgpt(query, dataset_id):
    """curl --location --request POST 'https://api.fastgpt.in/api/core/dataset/searchTest' \
    --header 'Authorization: Bearer fastgpt-xxxxx' \
    --header 'Content-Type: application/json' \
    --data-raw '{
        "datasetId": "知识库的ID",
        "text": "导演是谁",
        "limit": 5000,
        "similarity": 0,
        "searchMode": "embedding",
        "usingReRank": false
    }'
    """
    url = 'https://api.fastgpt.in/api/core/dataset/searchTest'
    headers = {
        'Authorization': f'Bearer {FAST_GPT_API_KEY}',
        'Content-Type': 'application/json',
    }
    data = {
        "datasetId": dataset_id,
        "text": query,
        "limit": 5000,
        "similarity": 0,
        "searchMode": "embedding",
        "usingReRank": True
    }
    response = requests.post(url, headers=headers, data=json.dumps(data)).json()
    print(response)
    return response             
    
def question_answer(excel_file):
    df = pd.read_excel(excel_file)
    all_questions_info = list()
    # 遍历每一行
    for index, row in df.iterrows():
        # 遍历每一列
        question_info = dict()
        for column in df.columns:
            value = row[column]
            question_info[column] = value
            # print(f"Row {index}, Column {column} has value {value}")
        all_questions_info.append(question_info)  

    for questions_info in tqdm(all_questions_info):
        question = questions_info['问题']
        file_type = questions_info['文件类型']
        collection_name = questions_info['知识库名']
        dataset_id = questions_info['dataset_id']
        if collection_name not in datasets or dataset_id == '':
            res = create_dataset(collection_name)
            dataset_id = add_dataset(collection_name, res['id']) 
        
        try:
            re_config_dataset(dataset_id)
            # response = search_dataset_from_fastgpt(question, dataset_id)
            response = get_answer_from_fastgpt(question)
            ans = response['choices'][0]['message']['content']
        except Exception as e:
            print(f'error in query {question}')
            ans = ''    
        print('ans:', ans, 'question:', question)
        questions_info['fastgpt_answer'] = ans
    
        df = pd.DataFrame(all_questions_info)
        df.to_excel(os.path.join('questions_info_with_answer_fastgpt.xlsx'), index=False)

if __name__ == '__main__':
    data_dir = './rag_benchmark_v1.0/rag_benchmark'
    save_dir = './rag_benchmark_v1.0/rag_benchmark_processed'
    excel_file = './data/questions_info_with_answer_sample.xlsx'
    save_excel_file = './data/questions_info_with_fastgpt_file_id.xlsx'
    
    query_dataset_from_fastgpt()
    # upload_file_to_fastgpt(save_dir, excel_file, save_excel_file)
    question_answer(save_excel_file)
    



# flake8: noqa
import io
import json
import logging
import os
import random
import tempfile
import time
from abc import ABC
from copy import deepcopy
from pathlib import Path
from typing import Any, Iterator, List, Mapping, Optional, Union
from urllib.parse import urlparse

import cv2
import fitz
import numpy as np
import pypdfium2
import requests
from image import LayoutParser
from langchain.document_loaders.blob_loaders import Blob


def norm_rect(bbox):
    x0 = np.min([bbox[0], bbox[2]])
    x1 = np.max([bbox[0], bbox[2]])
    y0 = np.min([bbox[1], bbox[3]])
    y1 = np.max([bbox[1], bbox[3]])
    return np.asarray([x0, y0, x1, y1])


def merge_rects(bboxes):
    x0 = np.min(bboxes[:, 0])
    y0 = np.min(bboxes[:, 1])
    x1 = np.max(bboxes[:, 2])
    y1 = np.max(bboxes[:, 3])
    return [x0, y0, x1, y1]


def get_image_blobs(pages, pdf_reader, n, start=0):
    blobs = []
    for pg in range(start, start + n):
        bytes_img = None
        page = pages.load_page(pg)
        mat = fitz.Matrix(1, 1)
        try:
            pm = page.get_pixmap(matrix=mat, alpha=False)
            bytes_img = pm.getPNGData()
        except Exception:
            # some pdf input cannot get render image from fitz
            page = pdf_reader.get_page(pg)
            pil_image = page.render().to_pil()
            img_byte_arr = io.BytesIO()
            pil_image.save(img_byte_arr, format='PNG')
            bytes_img = img_byte_arr.getvalue()

            blobs.append(Blob(data=bytes_img))
    return blobs


def test():
    file_path = './data/达梦数据库招股说明书_test_v1.pdf'
    blob = Blob.from_path(file_path)
    pages = None
    image_blobs = []
    with blob.as_bytes_io() as file_path:
        pages = fitz.open(file_path)
        pdf_reader = pypdfium2.PdfDocument(file_path, autoclose=True)
        image_blobs = get_image_blobs(pages, pdf_reader)

    assert len(image_blobs) == pages.page_count
    layout = LayoutParser()
    res = layout.parse(image_blobs[0])


def draw_polygon(image, bbox, text=None, color=(255, 0, 0), thickness=1):
    bbox = bbox.astype(np.int32)
    is_rect = bbox.shape[0] == 4
    if is_rect:
        start_point = (bbox[0], bbox[1])
        end_point = (bbox[2], bbox[3])
        image = cv2.rectangle(image, start_point, end_point, color, thickness)
    else:
        polys = [bbox.astype(np.int32).reshape((-1, 1, 2))]
        cv2.polylines(image, polys, True, color=color, thickness=thickness)
        start_point = (polys[0][0, 0, 0], polys[0][0, 0, 1])

    if text:
        fontFace = cv2.FONT_HERSHEY_SIMPLEX
        fontScale = 0.5
        color = (0, 0, 255)
        image = cv2.putText(image, text, start_point, fontFace, fontScale,
                            color, 1)

    return image


def test_vis():
    # file_path = './data/达梦数据库招股说明书_test_v1.pdf'
    file_path = './data/pdf_input/《中国药典》2020年版 一部.pdf'
    output_prefix = 'zhongguoyaodian_2020_v1'
    start, end, n = 70, 80, 10
    blob = Blob.from_path(file_path)
    pages = None
    image_blobs = []
    with blob.as_bytes_io() as file_path:
        pages = fitz.open(file_path)
        pdf_reader = pypdfium2.PdfDocument(file_path, autoclose=True)
        image_blobs = get_image_blobs(pages, pdf_reader, n, start)

    assert len(image_blobs) == n

    for i, blob in enumerate(image_blobs):
        idx = i + start
        # blob = image_blobs[2]
        layout = LayoutParser()
        out = layout.parse(blob)
        res = json.loads(out[0].page_content)
        bboxes = []
        labels = []
        for r in res:
            bboxes.append(r['bbox'])
            labels.append(str(r['category_id']))

        bboxes = np.asarray(bboxes)

        bytes_arr = np.frombuffer(blob.as_bytes(), dtype=np.uint8)
        image = cv2.imdecode(bytes_arr, flags=1)
        for bbox, text in zip(bboxes, labels):
            image = draw_polygon(image, bbox, text)

        outf = f'./data/{output_prefix}_layout_p{idx+1}_vis.png'
        cv2.imwrite(outf, image)


def order_by_tbyx(block_info, th=10):
    """
      block_info: [(b0, b1, b2, b3, text, x, y)+]
      th: threshold of the position threshold
    """
    # sort using y1 first and then x1
    res = sorted(block_info, key=lambda b: (b[1], b[0]))
    for i in range(len(res) - 1):
        for j in range(i, 0, -1):
            # restore the order using the
            if (abs(res[j + 1][1] - res[j][1]) < th
                    and (res[j + 1][0] < res[j][0])):
                tmp = deepcopy(res[j])
                res[j] = deepcopy(res[j + 1])
                res[j + 1] = deepcopy(tmp)
            else:
                break
    return res


def test_vis2():
    # file_path = './data/达梦数据库招股说明书_test_v1.pdf'
    file_path = './data/pdf_input/达梦数据库招股说明书.pdf'
    output_prefix = 'dameng_pageblock'

    start = 0
    end = 10
    n = end - start
    blob = Blob.from_path(file_path)
    pages = None
    image_blobs = []
    with blob.as_bytes_io() as file_path:
        pages = fitz.open(file_path)
        pdf_reader = pypdfium2.PdfDocument(file_path, autoclose=True)
        image_blobs = get_image_blobs(pages, pdf_reader, n, start)

    assert len(image_blobs) == pages.page_count

    for i, blob in enumerate(image_blobs):
        idx = i + start
        page = pages.load_page(idx)

        rect = page.rect
        print('rect', rect)
        o = 10
        b0 = np.asarray([rect.x0 + o, rect.y0 + o, rect.x1 - o, rect.y1 - o])

        bytes_arr = np.frombuffer(blob.as_bytes(), dtype=np.uint8)
        image = cv2.imdecode(bytes_arr, flags=1)

        image = draw_polygon(image, b0, '0.0')

        textpage = page.get_textpage()
        blocks = textpage.extractBLOCKS()
        IMG_BLOCK_TYPE = 1

        # blocks = order_by_tbyx(blocks)
        bboxes = []
        for off, b in enumerate(blocks):
            label = 'text' if b[-1] != IMG_BLOCK_TYPE else 'image'
            label = f'{label}-{off}'
            print('block', b, label)
            bbox = np.asarray([b[0], b[1], b[2], b[3]])
            bboxes.append(bbox)

            image = draw_polygon(image, bbox, label)

        if bboxes:
            b1 = merge_rects(np.asarray(bboxes))
            b1 = np.asarray(b1)
            image = draw_polygon(image, b1, '0.1')

        outf = f'./data/{output_prefix}_p{idx}_vis.png'
        cv2.imwrite(outf, image)


def test_vis3():
    file_path = './data/pdf_input/《中国药典》2020年版 一部.pdf'

    start = 50
    end = 60
    n = end - start
    output_prefix = 'zhongguoyaodian_2020_v1'

    blob = Blob.from_path(file_path)
    pages = None
    image_blobs = []
    with blob.as_bytes_io() as file_path:
        pages = fitz.open(file_path)
        pdf_reader = pypdfium2.PdfDocument(file_path, autoclose=True)
        image_blobs = get_image_blobs(pages, pdf_reader, n, start=50)

    assert len(image_blobs) == n

    for i, blob in enumerate(image_blobs):
        idx = i + start
        page = pages.load_page(idx)

        rect = page.rect
        print('rect', rect)
        o = 10
        b0 = np.asarray([rect.x0 + o, rect.y0 + o, rect.x1 - o, rect.y1 - o])

        bytes_arr = np.frombuffer(blob.as_bytes(), dtype=np.uint8)
        image = cv2.imdecode(bytes_arr, flags=1)

        image = draw_polygon(image, b0, '0.0')

        rotation_matrix = np.asarray(page.rotation_matrix).reshape((3, 2))
        c1 = (rotation_matrix[0, 0] - 1) <= 1e-6
        c2 = (rotation_matrix[1, 1] - 1) <= 1e-6
        is_rotated = c1 and c2

        textpage = page.get_textpage()
        blocks = textpage.extractBLOCKS()
        IMG_BLOCK_TYPE = 1

        # blocks = order_by_tbyx(blocks)
        bboxes = []
        for off, b in enumerate(blocks):
            label = 'text' if b[-1] != IMG_BLOCK_TYPE else 'image'
            label = f'{label}-{off}'
            print('block', b, label)
            bbox = np.asarray([b[0], b[1], b[2], b[3]])

            aug_bbox = bbox.reshape((-1, 2))
            padding = np.ones((len(aug_bbox), 1))
            aug_bbox = np.hstack([aug_bbox, padding])
            new_bbox = np.dot(aug_bbox, rotation_matrix).reshape(-1)

            new_bbox = norm_rect(new_bbox)

            print('new_bboxes', new_bbox)
            bboxes.append(new_bbox)

            image = draw_polygon(image, new_bbox, label)

        print(bboxes)
        if bboxes:
            b1 = merge_rects(np.asarray(bboxes))
            b1 = np.asarray(b1)
            image = draw_polygon(image, b1, '0.1')

        outf = f'./data/{output_prefix}_p{idx}_vis.png'
        cv2.imwrite(outf, image)


# test_vis3()
# test_vis2()
test_vis()
# test()



from typing import Any, Awaitable, Callable, Dict, List, Optional, Union

from autogen import Agent, ConversableAgent


class AutoGenCustomRole(ConversableAgent):
    """Custom agent that can use langchain agent and chain."""

    def __init__(
        self,
        name: str,
        system_message: str,
        func: Callable[..., str],
        coroutine: Optional[Callable[..., Awaitable[str]]] = None,
        **kwargs,
    ):
        super().__init__(name=name,
                         system_message=system_message,
                         human_input_mode='NEVER',
                         code_execution_config=False,
                         llm_config=False,
                         **kwargs)
        self.func = func
        self.coroutine = coroutine
        self.register_reply(Agent, AutoGenCustomRole.generate_custom_reply)
        self.register_reply(Agent, AutoGenCustomRole.a_generate_custom_reply, is_async=True)

    def generate_custom_reply(
        self,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Agent] = None,
        config: Optional[Any] = None,
        **kwargs,
    ) -> Union[str, Dict, None]:
        if messages is None:
            messages = self._oai_messages[sender]
        message = messages[-1]

        if 'content' in message:
            query = message['content']
            reply = self.func(query)
            if isinstance(reply, dict):
                reply = list(reply.values())
                if reply:
                    reply = str(reply[0])
                else:
                    reply = ''
            return True, reply

        return False, None

    async def a_generate_custom_reply(
        self,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Agent] = None,
        config: Optional[Any] = None,
        **kwargs,
    ) -> Union[str, Dict, None]:
        if messages is None:
            messages = self._oai_messages[sender]
        message = messages[-1]

        if 'content' in message:
            query = message['content']
            if self.coroutine:
                reply = await self.coroutine(query)
            else:
                reply = self.func(query)
            if isinstance(reply, dict):
                reply = list(reply.values())
                if reply:
                    reply = str(reply[0])
                else:
                    reply = ''
            return True, reply

        return False, None



# flake8: noqa
"""Loads PDF with semantic splilter."""
import io
import json
import os
import re
import time
from collections import Counter
from copy import deepcopy
from typing import List, Optional, Union

import fitz
import numpy as np
from bisheng_langchain.document_loaders.parsers import LayoutParser
from langchain.docstore.document import Document
from langchain.document_loaders.blob_loaders import Blob
from langchain.document_loaders.pdf import BasePDFLoader
from shapely import Polygon
from shapely import box as Rect

RE_MULTISPACE_INCLUDING_NEWLINES = re.compile(pattern=r'\s+', flags=re.DOTALL)


def merge_rects(bboxes):
    x0 = np.min(bboxes[:, 0])
    y0 = np.min(bboxes[:, 1])
    x1 = np.max(bboxes[:, 2])
    y1 = np.max(bboxes[:, 3])
    return [x0, y0, x1, y1]


def norm_rect(bbox):
    x0 = np.min([bbox[0], bbox[2]])
    x1 = np.max([bbox[0], bbox[2]])
    y0 = np.min([bbox[1], bbox[3]])
    y1 = np.max([bbox[1], bbox[3]])
    return np.asarray([x0, y0, x1, y1])


def find_max_continuous_seq(arr):
    n = len(arr)
    max_info = (0, 1)
    for i in range(n):
        m = 1
        for j in range(i + 1, n):
            if arr[j] - arr[j - 1] == 1:
                m += 1
            else:
                break

        if m > max_info[1]:
            max_info = (i, m)

    max_info = (max_info[0] + arr[0], max_info[1])
    return max_info


def order_by_tbyx(block_info, th=10):
    """
      block_info: [(b0, b1, b2, b3, text, x, y)+]
      th: threshold of the position threshold
    """
    # sort using y1 first and then x1
    res = sorted(block_info, key=lambda b: (b[1], b[0]))
    for i in range(len(res) - 1):
        for j in range(i, 0, -1):
            # restore the order using the
            if (abs(res[j + 1][1] - res[j][1]) < th and (res[j + 1][0] < res[j][0])):
                tmp = deepcopy(res[j])
                res[j] = deepcopy(res[j + 1])
                res[j + 1] = deepcopy(tmp)
            else:
                break
    return res


def join_lines(texts, is_table=False):
    if is_table:
        return '\n'.join(texts)

    flags = []
    PUNC_SET = set(['.', ',', ';', '?', '!'])
    for text in texts:
        flags.append(np.all([t.isalnum() for t in text.rsplit(' ', 5)]))

    if np.all(flags):
        t0 = texts[0]
        for t in texts[1:]:
            if t0[-1] == '-':
                t0 = t0[:-1] + t
            elif t0[-1].isalnum() and t[0].isalnum():
                t0 += ' ' + t
            elif t0[-1] in PUNC_SET or t[0] in PUNC_SET:
                t0 += ' ' + t
            else:
                t0 += t
        return t0
    else:
        return ''.join(texts)


class Segment:

    def __init__(self, seg):
        self.whole = seg
        self.segs = []

    @staticmethod
    def is_align(seg0, seg1, delta=5, mode=0):
        # mode=0 edge align
        # mode=1, edge align or center align
        res = Segment.contain(seg0, seg1)
        if not res:
            return False
        else:
            if mode == 1:
                r1 = seg1[0] - seg0[0] <= delta or seg0[1] - seg1[1] <= delta
                c0 = (seg0[0] + seg0[1]) / 2
                c1 = (seg1[0] + seg1[1]) / 2
                r2 = abs(c1 - c0) <= delta
                return r1 or r2
            else:
                return seg1[0] - seg0[0] <= delta or seg0[1] - seg1[1] <= delta

    @staticmethod
    def contain(seg0, seg1):
        return seg0[0] <= seg1[0] and seg0[1] >= seg1[0]

    @staticmethod
    def overlap(seg0, seg1):
        max_x0 = max(seg0[0], seg1[0])
        min_x1 = min(seg0[1], seg1[1])
        return max_x0 < min_x1

    def _merge(self, segs):
        x0s = [s[0] for s in segs]
        x1s = [s[1] for s in segs]
        return (np.min(x0s), np.max(x1s))

    def add(self, seg):
        if not self.segs:
            self.segs.append(seg)
        else:
            overlaps = []
            non_overlaps = []
            for seg0 in self.segs:
                if Segment.overlap(seg0, seg):
                    overlaps.append(seg0)
                else:
                    non_overlaps.append(seg0)

            if not overlaps:
                self.segs.append(seg)
            else:
                overlaps.append(seg)
                new_seg = self._merge(overlaps)
                non_overlaps.append(new_seg)
                self.segs = non_overlaps

    def get_free_segment(self, incr_margin=True, margin_threshold=10):
        sorted_segs = sorted(self.segs, key=lambda x: x[0])
        n = len(sorted_segs)
        free_segs = []
        if incr_margin:
            if n > 0:
                seg_1st = sorted_segs[0]
                if (seg_1st[0] - self.whole[0]) > margin_threshold:
                    free_segs.append((self.whole[0], seg_1st[0]))

                seg_last = sorted_segs[-1]
                if (self.whole[1] - seg_last[1]) > margin_threshold:
                    free_segs.append((seg_last[1], self.whole[1]))

        for i in range(n - 1):
            x0 = sorted_segs[i][1]
            x1 = sorted_segs[i + 1][0]
            free_segs.append((x0, x1))

        return free_segs


class PDFWithSemanticLoader(BasePDFLoader):
    """Loads a PDF with pypdf and chunks at character level.

    Loader also stores page numbers in metadata.
    """

    def __init__(self,
                 file_path: str,
                 password: Optional[Union[str, bytes]] = None,
                 layout_api_key: str = None,
                 layout_api_url: str = None,
                 is_join_table: bool = True,
                 with_columns: bool = False,
                 support_rotate: bool = False,
                 text_elem_sep: str = '\n',
                 start: int = 0,
                 n: int = None,
                 html_output_file: str = None,
                 verbose: bool = False) -> None:
        """Initialize with a file path."""
        self.layout_parser = LayoutParser(api_key=layout_api_key, api_base_url=layout_api_url)
        self.with_columns = with_columns
        self.is_join_table = is_join_table
        self.support_rotate = support_rotate
        self.start = start
        self.n = n
        self.html_output_file = html_output_file
        self.verbose = verbose
        self.text_elem_sep = text_elem_sep
        super().__init__(file_path)

    def _get_image_blobs(self, fitz_doc, pdf_reader, n=None, start=0):
        blobs = []
        pages = []
        if not n:
            n = fitz_doc.page_count
        for pg in range(start, start + n):
            bytes_img = None
            page = fitz_doc.load_page(pg)
            pages.append(page)
            mat = fitz.Matrix(1, 1)
            try:
                pm = page.get_pixmap(matrix=mat, alpha=False)
                bytes_img = pm.getPNGData()
            except Exception:
                # some pdf input cannot get render image from fitz
                page = pdf_reader.get_page(pg)
                pil_image = page.render().to_pil()
                img_byte_arr = io.BytesIO()
                pil_image.save(img_byte_arr, format='PNG')
                bytes_img = img_byte_arr.getvalue()

            blobs.append(Blob(data=bytes_img))
        return blobs, pages

    def _allocate_semantic(self, page, layout):
        class_name = ['印章', '图片', '标题', '段落', '表格', '页眉', '页码', '页脚']
        effective_class_inds = [3, 4, 5, 999]
        non_conti_class_ids = [6, 7, 8]
        TEXT_ID = 4
        TABLE_ID = 5

        textpage = page.get_textpage()
        blocks = textpage.extractBLOCKS()

        if self.support_rotate:
            rotation_matrix = np.asarray(page.rotation_matrix).reshape((3, 2))
            c1 = (rotation_matrix[0, 0] - 1) <= 1e-6
            c2 = (rotation_matrix[1, 1] - 1) <= 1e-6
            is_rotated = c1 and c2
            # print('c1/c2', c1, c2)
            if is_rotated:
                new_blocks = []
                for b in blocks:
                    bbox = np.asarray([b[0], b[1], b[2], b[3]])
                    aug_bbox = bbox.reshape((-1, 2))
                    padding = np.ones((len(aug_bbox), 1))
                    aug_bbox = np.hstack([aug_bbox, padding])
                    bb = np.dot(aug_bbox, rotation_matrix).reshape(-1)
                    bb = norm_rect(bb)
                    info = (bb[0], bb[1], bb[2], bb[3], b[4], b[5], b[6])
                    new_blocks.append(info)

                blocks = new_blocks

        if not self.with_columns:
            blocks = order_by_tbyx(blocks)

        # print('---ori blocks---')
        # for b in blocks:
        #     print(b)

        IMG_BLOCK_TYPE = 1
        text_ploys = []
        text_rects = []
        texts = []
        for b in blocks:
            if b[-1] != IMG_BLOCK_TYPE:
                text = re.sub(RE_MULTISPACE_INCLUDING_NEWLINES, ' ', b[4] or '').strip()
                if text:
                    texts.append(text)
                    text_ploys.append(Rect(b[0], b[1], b[2], b[3]))
                    text_rects.append([b[0], b[1], b[2], b[3]])
        text_rects = np.asarray(text_rects)
        texts = np.asarray(texts)

        semantic_polys = []
        semantic_labels = []

        layout_info = json.loads(layout.page_content)
        for info in layout_info:
            bbs = info['bbox']
            coords = ((bbs[0], bbs[1]), (bbs[2], bbs[3]), (bbs[4], bbs[5]), (bbs[6], bbs[7]))
            semantic_polys.append(Polygon(coords))
            semantic_labels.append(info['category_id'])

        # caculate containing overlap
        sem_cnt = len(semantic_polys)
        texts_cnt = len(text_ploys)
        contain_matrix = np.zeros((sem_cnt, texts_cnt))
        for i in range(sem_cnt):
            for j in range(texts_cnt):
                inter = semantic_polys[i].intersection(text_ploys[j]).area
                contain_matrix[i, j] = inter * 1.0 / text_ploys[j].area

        # print('----------------containing matrix--------')
        # for r in contain_matrix.tolist():
        #     print([round(r_, 2) for r_ in r])

        # print('---text---')
        # for t in texts:
        #     print(t)

        # merge continuous text block by the containing matrix
        CONTRAIN_THRESHOLD = 0.70
        contain_info = []
        for i in range(sem_cnt):
            ind = np.argwhere(contain_matrix[i, :] > CONTRAIN_THRESHOLD)[:, 0]
            if len(ind) == 0: continue
            label = semantic_labels[i]
            if label in non_conti_class_ids:
                n = len(ind)
                contain_info.append((None, None, n, label, ind))
            else:
                start, n = find_max_continuous_seq(ind)
                if n >= 1:
                    contain_info.append((start, start + n, n, label, None))

        contain_info = sorted(contain_info, key=lambda x: x[2], reverse=True)
        mask = np.zeros(texts_cnt)
        new_block_info = []
        for info in contain_info:
            start, end, n, label, ind = info
            if label in non_conti_class_ids and np.all(mask[ind] == 0):
                rect = merge_rects(text_rects[ind])
                ori_orders = [blocks[i][-2] for i in ind]
                ts = texts[ind]
                rs = text_rects[ind]
                ord_ind = np.min(ori_orders)
                mask[ind] = 1
                new_block_info.append((rect[0], rect[1], rect[2], rect[3], ts, rs, ord_ind))

            elif np.all(mask[start:end] == 0):
                rect = merge_rects(text_rects[start:end])
                ori_orders = [blocks[i][-2] for i in range(start, end)]
                arg_ind = np.argsort(ori_orders)
                # print('ori_orders', ori_orders, arg_ind)
                ord_ind = np.min(ori_orders)

                ts = texts[start:end]
                rs = text_rects[start:end]
                if label == TABLE_ID:
                    ts = ts[arg_ind]
                    rs = rs[arg_ind]

                mask[start:end] = 1
                new_block_info.append((rect[0], rect[1], rect[2], rect[3], ts, rs, ord_ind))

        for i in range(texts_cnt):
            if mask[i] == 0:
                b = blocks[i]
                r = np.asarray([b[0], b[1], b[2], b[3]])
                ord_ind = b[-2]
                new_block_info.append((b[0], b[1], b[2], b[3], [texts[i]], [r], ord_ind))

        if self.with_columns:
            new_blocks = sorted(new_block_info, key=lambda x: x[-1])
        else:
            new_blocks = order_by_tbyx(new_block_info)

        # print('\n\n---new blocks---')
        # for idx, b in enumerate(new_blocks):
        #     print(idx, b)

        text_ploys = []
        texts = []
        for b in new_blocks:
            texts.append(b[4])
            text_ploys.append(Rect(b[0], b[1], b[2], b[3]))

        # caculate overlap
        sem_cnt = len(semantic_polys)
        texts_cnt = len(text_ploys)
        overlap_matrix = np.zeros((sem_cnt, texts_cnt))
        for i in range(sem_cnt):
            for j in range(texts_cnt):
                inter = semantic_polys[i].intersection(text_ploys[j]).area
                union = semantic_polys[i].union(text_ploys[j]).area
                overlap_matrix[i, j] = (inter * 1.0) / union

        # print('---overlap_matrix---')
        # for r in overlap_matrix:
        #     print([round(r_, 3) for r_ in r])

        # allocate label
        OVERLAP_THRESHOLD = 0.2
        texts_labels = []
        DEF_SEM_LABEL = 999
        for j in range(texts_cnt):
            ind = np.argwhere(overlap_matrix[:, j] > OVERLAP_THRESHOLD)[:, 0]
            if len(ind) == 0:
                sem_label = DEF_SEM_LABEL
            else:
                c = Counter([semantic_labels[i] for i in ind])
                items = c.most_common()
                sem_label = items[0][0]
                if len(items) > 1 and TEXT_ID in dict(items):
                    sem_label = TEXT_ID

            texts_labels.append(sem_label)

        # print(texts_labels)
        # filter the unused element
        filtered_blocks = []
        for label, b in zip(texts_labels, new_blocks):
            if label in effective_class_inds:
                text = join_lines(b[4], label == TABLE_ID)
                filtered_blocks.append((b[0], b[1], b[2], b[3], text, b[5], label))

        # print('---filtered_blocks---')
        # for b in filtered_blocks:
        #     print(b)

        return filtered_blocks

    def _divide_blocks_into_groups(self, blocks):
        # support only pure two columns layout, each has same width
        rects = np.asarray([[b[0], b[1], b[2], b[3]] for b in blocks])
        min_x0 = np.min(rects[:, 0])
        max_x1 = np.max(rects[:, 2])
        root_seg = (min_x0, max_x1)
        root_pc = (min_x0 + max_x1) / 2
        root_offset = 20
        center_seg = (root_pc - root_offset, root_pc + root_offset)

        segment = Segment(root_seg)
        for r in rects:
            segment.add((r[0], r[2]))

        COLUMN_THRESHOLD = 0.90
        CENTER_GAP_THRESHOLD = 0.90
        free_segs = segment.get_free_segment()
        columns = []
        if len(free_segs) == 1 and len(segment.segs) == 2:
            free_seg = free_segs[0]
            seg0 = segment.segs[0]
            seg1 = segment.segs[1]
            cover = seg0[1] - seg0[0] + seg1[1] - seg1[0]
            c0 = cover / (root_seg[1] - root_seg[0])
            c1 = Segment.contain(center_seg, free_seg)
            if c0 > COLUMN_THRESHOLD and c1:
                # two columns
                columns.extend([seg0, seg1])

        groups = [blocks]
        if columns:
            groups = [[] for _ in columns]
            for b, r in zip(blocks, rects):
                column_ind = 0
                cand_seg = (r[0], r[2])
                for i, seg in enumerate(columns):
                    if Segment.contain(seg, cand_seg):
                        column_ind = i
                        break
                groups[i].append(b)

        return groups

    def _allocate_continuous(self, groups):
        g_bound = []
        groups = [g for g in groups if g]
        for blocks in groups:
            arr = [[b[0], b[1], b[2], b[3]] for b in blocks]
            bboxes = np.asarray(arr)
            g_bound.append(np.asarray(merge_rects(bboxes)))

        LINE_FULL_THRESHOLD = 0.80
        START_THRESHOLD = 0.8
        SIMI_HEIGHT_THRESHOLD = 0.3
        SIMI_WIDTH_THRESHOLD = 0.3

        TEXT_ID = 4
        TABLE_ID = 5

        def _get_elem(blocks, is_first=True):
            if not blocks:
                return (None, None, None, None, None)
            if is_first:
                b1 = blocks[0]
                b1_label = b1[-1]
                r1 = b1[5][0]
                r1_w = r1[2] - r1[0]
                r1_h = r1[3] - r1[1]
                return (b1, b1_label, r1, r1_w, r1_h)
            else:
                b0 = blocks[-1]
                b0_label = b0[-1]
                r0 = b0[5][-1]
                r0_w = r0[2] - r0[0]
                r0_h = r0[3] - r0[1]
                return (b0, b0_label, r0, r0_w, r0_h)

        b0, b0_label, r0, r0_w, r0_h = _get_elem(groups[0], False)
        g0 = g_bound[0]

        for i in range(1, len(groups)):
            b1, b1_label, r1, r1_w, r1_h = _get_elem(groups[i], True)
            g1 = g_bound[i]

            # print('\n_allocate_continuous:')
            # print(b0, b0_label, b1, b1_label)

            if b0_label and b0_label == b1_label and b0_label == TEXT_ID:
                c0 = r0_w / (g0[2] - g0[0])
                c1 = (r1[0] - g1[0]) / r1_h
                c2 = np.abs(r0_h - r1_h) / r1_h

                # print('\n\n---conti texts---')
                # print(b0_label, c0, c1, c2,
                #       b0, b0_label, r0, r0_w, r0_h,
                #       b1, b1_label, r1, r1_w, r1_h)

                if (c0 > LINE_FULL_THRESHOLD and c1 < START_THRESHOLD
                        and c2 < SIMI_HEIGHT_THRESHOLD):
                    new_text = join_lines([b0[4], b1[4]])
                    new_block = (b0[0], b0[1], b0[2], b0[3], new_text, b0[5], b0[6])
                    groups[i - 1][-1] = new_block
                    groups[i].pop(0)

            elif (self.is_join_table and b0_label and b0_label == b1_label
                  and b0_label == TABLE_ID):
                c0 = (r1_w - r0_w) / r1_h
                if c0 < SIMI_WIDTH_THRESHOLD:
                    new_text = join_lines([b0[4], b1[4]], True)
                    new_block = (b0[0], b0[1], b0[2], b0[3], new_text, b0[5], b0[6])
                    groups[i - 1][-1] = new_block
                    groups[i].pop(0)

            b0, b0_label, r0, r0_w, r0_h = _get_elem(groups[i], False)

        return groups

    def save_to_html(self, groups, output_file):
        styles = ['style="background-color: #EBEBEB;"', 'style="background-color: #ABBAEA;"']
        idx = 0
        table_style = 'style="border:1px solid black;"'

        with open(output_file, 'w') as fout:
            for blocks in groups:
                for b in blocks:
                    if b[-1] == 3:
                        text = f'<h1>{b[4]}</h1>'
                    elif b[-1] == 4:
                        text = f'<p {styles[idx % 2]}>{b[4]}</p>'
                        idx += 1
                    elif b[-1] == 5:
                        rows = b[4].split('\n')
                        content = []
                        for r in rows:
                            content.append(f'<tr><td {table_style}>{r}</td></tr>')
                        elem_text = '\n'.join(content)
                        text = f'<table {table_style}>{elem_text}</table>'
                    else:
                        text = f'<p {styles[idx % 2]}>{b[4]}</p>'
                        idx += 1

                    fout.write(text + '\n')

    def _save_to_document(self, groups):
        TITLE_ID = 3
        TEXT_ID = 4
        TABLE_ID = 5
        content_page = []
        is_first_elem = True
        for blocks in groups:
            for b in blocks:
                if is_first_elem:
                    content_page.append(b[4])
                    is_first_elem = False
                else:
                    label, text = b[-1], b[4]
                    if label == TITLE_ID:
                        content_page.append('\n\n' + text)
                    else:
                        content_page.append(self.text_elem_sep + text)

        return ''.join(content_page)

    def load(self) -> List[Document]:
        """Load given path as pages."""
        import pypdfium2
        blob = Blob.from_path(self.file_path)
        start = self.start
        groups = []
        with blob.as_bytes_io() as file_path:
            fitz_doc = fitz.open(file_path)
            pdf_doc = pypdfium2.PdfDocument(file_path, autoclose=True)
            max_page = fitz_doc.page_count - start
            n = self.n if self.n else max_page
            n = min(n, max_page)

            tic = time.time()
            if self.verbose:
                print(f'{n} pages need be processed...')

            for idx in range(start, start + n):
                blobs, pages = self._get_image_blobs(fitz_doc, pdf_doc, 1, idx)
                layout = self.layout_parser.parse(blobs[0])[0]
                blocks = self._allocate_semantic(pages[0], layout)
                if not blocks: continue

                if self.with_columns:
                    sub_groups = self._divide_blocks_into_groups(blocks)
                    groups.extend(sub_groups)
                else:
                    groups.append(blocks)

                if self.verbose:
                    count = idx - start + 1
                    if count % 50 == 0:
                        elapse = round(time.time() - tic, 2)
                        tic = time.time()
                        print(f'process {count} pages used {elapse}sec...')

        groups = self._allocate_continuous(groups)

        if self.html_output_file:
            self.save_to_html(groups, self.html_output_file)
            return []

        page_content = self._save_to_document(groups)
        meta = {'source': os.path.basename(self.file_path)}
        doc = Document(page_content=page_content, metadata=meta)
        return [doc]



from bisheng_langchain.agents.chatglm_functions_agent.base import ChatglmFunctionsAgent
from bisheng_langchain.agents.llm_functions_agent.base import LLMFunctionsAgent

__all__ = ['ChatglmFunctionsAgent', 'LLMFunctionsAgent']



import json
import os
import warnings
from typing import Any, Callable, Dict, List, Optional, Tuple

import httpx
import pandas as pd
import pymysql
from bisheng_langchain.gpts.tools.api_tools import ALL_API_TOOLS
from bisheng_langchain.gpts.tools.bing_search.tool import BingSearchRun
from bisheng_langchain.gpts.tools.calculator.tool import calculator
from bisheng_langchain.gpts.tools.code_interpreter.tool import CodeInterpreterTool

# from langchain_community.utilities.dalle_image_generator import DallEAPIWrapper
from bisheng_langchain.gpts.tools.dalle_image_generator.tool import (
    DallEAPIWrapper,
    DallEImageGenerator,
)
from bisheng_langchain.gpts.tools.get_current_time.tool import get_current_time
from dotenv import load_dotenv
from langchain_community.tools.arxiv.tool import ArxivQueryRun
from langchain_community.tools.bearly.tool import BearlyInterpreterTool
from langchain_community.utilities.arxiv import ArxivAPIWrapper
from langchain_community.utilities.bing_search import BingSearchAPIWrapper
from langchain_core.callbacks import BaseCallbackManager, Callbacks
from langchain_core.language_models import BaseLanguageModel
from langchain_core.tools import BaseTool, Tool
from mypy_extensions import Arg, KwArg
from bisheng_langchain.rag import BishengRAGTool
from bisheng_langchain.utils.azure_dalle_image_generator import AzureDallEWrapper


def _get_current_time() -> BaseTool:
    return get_current_time


def _get_calculator() -> BaseTool:
    return calculator


def _get_arxiv() -> BaseTool:
    return ArxivQueryRun(api_wrapper=ArxivAPIWrapper())


_BASE_TOOLS: Dict[str, Callable[[], BaseTool]] = {
    'get_current_time': _get_current_time,
    'calculator': _get_calculator,
    'arxiv': _get_arxiv,
}

_LLM_TOOLS: Dict[str, Callable[[BaseLanguageModel], BaseTool]] = {}

_EXTRA_LLM_TOOLS: Dict[
    str, Tuple[Callable[[Arg(BaseLanguageModel, 'llm'), KwArg(Any)], BaseTool], List[str]]  # noqa  # noqa #type: ignore
] = {}


def _get_bing_search(**kwargs: Any) -> BaseTool:
    return BingSearchRun(api_wrapper=BingSearchAPIWrapper(**kwargs))


def _get_dalle_image_generator(**kwargs: Any) -> Tool:
    if kwargs.get('openai_proxy'):
        kwargs['http_async_client'] = httpx.AsyncClient(proxies=kwargs.get('openai_proxy'))
        kwargs['http_client'] = httpx.Client(proxies=kwargs.get('openai_proxy'))

    # 说明是azure的openai配置
    if kwargs.get("azure_endpoint"):
        kwargs['api_key'] = kwargs.pop('openai_api_key')
        kwargs['api_version'] = kwargs.pop('openai_api_version')
        return DallEImageGenerator(
            api_wrapper=AzureDallEWrapper(**kwargs)
        )

    return DallEImageGenerator(
        api_wrapper=DallEAPIWrapper(
            model='dall-e-3',
            **kwargs
        )
    )


def _get_bearly_code_interpreter(**kwargs: Any) -> Tool:
    return BearlyInterpreterTool(**kwargs).as_tool()


def _get_native_code_interpreter(**kwargs: Any) -> Tool:
    return CodeInterpreterTool(**kwargs).as_tool()


# 第二个list内填必填参数，第三个list内填可选参数
_EXTRA_PARAM_TOOLS: Dict[str, Tuple[Callable[[KwArg(Any)], BaseTool], List[Optional[str]], List[Optional[str]]]] = {
    # type: ignore
    'dalle_image_generator': (_get_dalle_image_generator,
                              ['openai_api_key'],
                              ['openai_api_base', 'openai_proxy', 'azure_deployment', 'azure_endpoint', 'openai_api_version']),
    'bing_search': (_get_bing_search, ['bing_subscription_key', 'bing_search_url'], []),
    'bisheng_code_interpreter': (_get_native_code_interpreter, ["minio"], ['files']),
    'bisheng_rag': (BishengRAGTool.get_rag_tool, ['name', 'description'],
                    ['vector_store', 'keyword_store', 'llm', 'collection_name', 'max_content',
                     'sort_by_source_and_index']),
}

_API_TOOLS: Dict[str, Tuple[Callable[[KwArg(Any)], BaseTool], List[str]]] = {**ALL_API_TOOLS}  # type: ignore

_ALL_TOOLS = {
    **_BASE_TOOLS,
    **_LLM_TOOLS,
    **_EXTRA_LLM_TOOLS,
    **_EXTRA_PARAM_TOOLS,
    **_API_TOOLS,
}


def _handle_callbacks(callback_manager: Optional[BaseCallbackManager], callbacks: Callbacks) -> Callbacks:
    if callback_manager is not None:
        warnings.warn(
            'callback_manager is deprecated. Please use callbacks instead.',
            DeprecationWarning,
        )
        if callbacks is not None:
            raise ValueError('Cannot specify both callback_manager and callbacks arguments.')
        return callback_manager
    return callbacks


def load_tools(
        tool_params: Dict[str, Dict[str, Any]],
        llm: Optional[BaseLanguageModel] = None,
        callbacks: Callbacks = None,
        **kwargs: Any,
) -> List[BaseTool]:
    tools = []
    callbacks = _handle_callbacks(callback_manager=kwargs.get('callback_manager'), callbacks=callbacks)
    for name, params in tool_params.items():
        if name in _BASE_TOOLS:
            tools.append(_BASE_TOOLS[name]())
        elif name in _LLM_TOOLS:
            if llm is None:
                raise ValueError(f'Tool {name} requires an LLM to be provided')
            tool = _LLM_TOOLS[name](llm)
            tools.append(tool)
        elif name in _EXTRA_LLM_TOOLS:
            if llm is None:
                raise ValueError(f'Tool {name} requires an LLM to be provided')
            _get_llm_tool_func, extra_keys = _EXTRA_LLM_TOOLS[name]
            missing_keys = set(extra_keys).difference(params)
            if missing_keys:
                raise ValueError(f'Tool {name} requires some parameters that were not ' f'provided: {missing_keys}')
            sub_kwargs = {k: params[k] for k in extra_keys}
            tool = _get_llm_tool_func(llm=llm, **sub_kwargs)
            tools.append(tool)
        elif name in _EXTRA_PARAM_TOOLS:
            _get_tool_func, extra_keys, optional_keys = _EXTRA_PARAM_TOOLS[name]
            missing_keys = set(extra_keys).difference(params)
            if missing_keys:
                raise ValueError(f'Tool {name} requires some parameters that were not ' f'provided: {missing_keys}')
            extra_kwargs = {k: params[k] for k in extra_keys}
            optional_kwargs = {k: params[k] for k in optional_keys if k in params}
            all_kwargs = {**extra_kwargs, **optional_kwargs}
            tool = _get_tool_func(**all_kwargs)
            tools.append(tool)
        elif name in _API_TOOLS:
            _get_api_tool_func, extra_keys = _API_TOOLS[name]
            missing_keys = set(extra_keys).difference(params)
            if missing_keys:
                raise ValueError(f'Tool {name} requires some parameters that were not ' f'provided: {missing_keys}')
            mini_kwargs = {k: params[k] for k in extra_keys}
            tool = _get_api_tool_func(name=name, **mini_kwargs)
            tools.append(tool)
        else:
            raise ValueError(f'Got unknown tool {name}')
    if callbacks is not None:
        for tool in tools:
            tool.callbacks = callbacks
    return tools


def get_all_tool_names() -> List[str]:
    """Get a list of all possible tool names."""
    return list(_ALL_TOOLS.keys())


def get_tool_table():
    load_dotenv('.sql_env', override=True)
    db = pymysql.connect(
        host=os.getenv('MYSQL_HOST'),
        user=os.getenv('MYSQL_USER'),
        password=os.getenv('MYSQL_PASSWORD'),
        database=os.getenv('MYSQL_DATABASE'),
        port=int(os.getenv('MYSQL_PORT')),
    )
    cursor = db.cursor()
    cursor.execute("SELECT name, t.desc, tool_key, extra FROM t_gpts_tools as t;")
    results = cursor.fetchall()
    db.close()

    df = pd.DataFrame(
        columns=[
            '前端工具名',
            '前端工具描述',
            'tool_key',
            'tool参数配置',
            'function_name',
            'function_description',
            'function_args',
        ]
    )
    for i, result in enumerate(results):
        name, desc, tool_key, extra = result
        if not extra:
            extra = '{}'
        tool_func = load_tools({tool_key: json.loads(extra)})[0]

        df.loc[i, '前端工具名'] = name
        df.loc[i, '前端工具描述'] = desc
        df.loc[i, 'tool_key'] = tool_key
        df.loc[i, 'tool参数配置'] = extra
        df.loc[i, 'function_name'] = tool_func.name
        df.loc[i, 'function_description'] = tool_func.description
        df.loc[i, 'function_args'] = f"{tool_func.args_schema.schema()['properties']}"

    return df




from typing import List, Optional

from pydantic import BaseModel, Extra


class InputNode(BaseModel):
    """Input组件，用来控制输入"""
    input: Optional[List[str]]

    def text(self):
        return self.input


class VariableNode(BaseModel):
    """用来设置变量"""
    # key
    variables: Optional[List[str]]
    # vaulues
    variable_value: Optional[List[str]] = []

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    def text(self):
        if self.variable_value:
            text = {}
            for index, value in enumerate(self.variable_value):
                text[self.variables[index]] = value

            return text
        else:
            return {}


class InputFileNode(BaseModel):
    file_path: Optional[str]
    file_name: Optional[str]
    file_type: Optional[str]  # tips for file
    """Output组件，用来控制输出"""

    def text(self):
        # judge if file_path is oss address
        if not self.file_path:
            return ''
        return [self.file_path, self.file_name]



from loguru import logger
from langchain_core.pydantic_v1 import BaseModel, Field
from typing import Any
from .base import APIToolBase
from .base import MultArgsSchemaTool
from langchain_core.tools import BaseTool


class FlowTools(APIToolBase):

    def run(self, query: str) -> str:
        """Run query through api and parse result."""
        if query:
            self.params[self.input_key] = {
                'question': query
            }

        url = self.url
        logger.info('api_call url={}', url)
        resp = self.client.post(url, json=self.params)
        if resp.status_code != 200:
            logger.info('api_call_fail res={}', resp.text)
        return resp.text

    async def arun(self, query: str) -> str:
        """Run query through api and parse result."""
        if query:
            self.params[self.input_key] = {
                'question': query
            }

        url = self.url
        logger.info('api_call url={}', url)
        resp = await self.async_client.apost(url, json=self.params)
        logger.info(resp)
        return resp

    @classmethod
    def knowledge_retrieve(cls, collection_id: int = None) -> str:    

        flow_id = 'c7985115-a9d2-446a-9c55-40b5728ffb52'
        url = 'http://192.168.106.120:3002/api/v1/process/{}'.format(flow_id)
        input_key = 'inputs'
        params = {
            'tweaks': {
                'ElasticKeywordsSearch-pFFyR': {
                    'collection_id': collection_id
                },
                'Milvus-9KIR6': {
                    'collection_id': collection_id
                }
            }
        }

        class InputArgs(BaseModel):
            query: str = Field(description='questions to ask')

        return cls(url=url, params=params, input_key=input_key, args_schema=InputArgs)
    
    @classmethod
    def get_api_tool(cls, name, **kwargs: Any) -> BaseTool:
        attr_name = name.split('_', 1)[-1]
        class_method = getattr(cls, attr_name)
        function_description = kwargs.get('description','')
        kwargs.pop('description')
        return MultArgsSchemaTool(name=name + '_' +str(kwargs.get('collection_id')),
                                  description=function_description,
                                  func=class_method(**kwargs).run,
                                  coroutine=class_method(**kwargs).arun,
                                  args_schema=class_method(**kwargs).args_schema)
    



import json
import re
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from bisheng_langchain.chat_models.host_llm import HostChatGLM
from langchain.agents.agent import Agent, AgentOutputParser, BaseSingleActionAgent
from langchain.agents.structured_chat.output_parser import StructuredChatOutputParserWithRetries
from langchain.agents.structured_chat.prompt import FORMAT_INSTRUCTIONS, PREFIX, SUFFIX
from langchain.callbacks.base import BaseCallbackManager
from langchain.callbacks.manager import Callbacks
from langchain.prompts.chat import (ChatPromptTemplate, HumanMessagePromptTemplate,
                                    SystemMessagePromptTemplate)
from langchain.schema import AgentAction, AgentFinish, BasePromptTemplate
from langchain.schema.language_model import BaseLanguageModel
from langchain.schema.messages import ChatMessage
from langchain.tools import BaseTool, StructuredTool
from langchain_core.pydantic_v1 import Field, root_validator

HUMAN_MESSAGE_TEMPLATE = '{input}\n\n{agent_scratchpad}'


def format_tool_to_chatglm_function(tool: BaseTool):
    """Format tool into the chatglm function API."""
    if isinstance(tool, StructuredTool):
        schema_ = tool.args_schema.schema()
        # Bug with required missing for structured tools.
        required = sorted(schema_['properties'])  # BUG WORKAROUND
        return {
            'name': tool.name,
            'description': tool.description,
            'parameters': {
                'type': 'object',
                'properties': schema_['properties'],
                'required': required,
            },
        }
    else:
        if tool.args_schema:
            schema_ = tool.args_schema.schema()
            parameters = {
                'type': 'object',
                'properties': schema_['properties'],
                'required': list(schema_['properties'].keys()),
            }
        else:
            parameters = {
                # This is a hack to get around the fact that some tools
                # do not expose an args_schema, and expect an argument
                # which is a string.
                # And Open AI does not support an array type for the
                # parameters.
                'type': 'object',
                'properties': {
                    '__arg1': {
                        'type': 'string',
                        'title': '__arg1'
                    },
                },
                'required': ['__arg1'],
            }

        return {
            'name': tool.name,
            'description': tool.description,
            'parameters': parameters,
        }


class ChatglmFunctionsAgent(BaseSingleActionAgent):
    """Chatglm Functions Agent."""

    llm: BaseLanguageModel
    tools: Sequence[BaseTool]
    prompt: BasePromptTemplate
    output_parser: AgentOutputParser = Field(default_factory=StructuredChatOutputParserWithRetries)
    """Output parser for the agent."""
    has_search: bool = False
    history: List = []

    def get_allowed_tools(self) -> List[str]:
        """Get allowed tools."""
        return list([t.name for t in self.tools])

    @root_validator
    def validate_llm(cls, values: dict) -> dict:
        if not isinstance(values['llm'], HostChatGLM):
            raise ValueError('Only supported with ChatGLM3 models.')
        return values

    @root_validator
    def validate_prompt(cls, values: dict) -> dict:
        prompt: BasePromptTemplate = values['prompt']
        if 'agent_scratchpad' not in prompt.input_variables:
            raise ValueError('`agent_scratchpad` should be one of the variables in the prompt, '
                             f'got {prompt.input_variables}')
        return values

    @property
    def input_keys(self) -> List[str]:
        """Get input keys. Input refers to user input here."""
        return ['input']

    @property
    def observation_prefix(self) -> str:
        """Prefix to append the observation with."""
        return 'Observation: '

    @property
    def llm_prefix(self) -> str:
        """Prefix to append the llm call with."""
        return 'Thought:'

    def _construct_scratchpad(self, intermediate_steps: List[Tuple[AgentAction, str]]) -> str:
        agent_scratchpad = ''
        for action, observation in intermediate_steps:
            agent_scratchpad += action.log
            agent_scratchpad += f'\n{self.observation_prefix}{observation}\n{self.llm_prefix}'
        if not isinstance(agent_scratchpad, str):
            raise ValueError('agent_scratchpad should be of type string.')
        if agent_scratchpad:
            return (f'This was your previous work '
                    f"(but I haven't seen any of it! I only see what "
                    f'you return as final answer):\n{agent_scratchpad}')
        else:
            return agent_scratchpad

    @classmethod
    def _validate_tools(cls, tools: Sequence[BaseTool]) -> None:
        pass

    @classmethod
    def _get_default_output_parser(cls,
                                   llm: Optional[BaseLanguageModel] = None,
                                   **kwargs: Any) -> AgentOutputParser:
        return StructuredChatOutputParserWithRetries.from_llm(llm=llm)

    @property
    def _stop(self) -> List[str]:
        return ['Observation:']

    def _tool_history(self, prompt: str):
        ans = []
        tools_json = []
        for i, tool in enumerate(self.tools):
            tool_config = format_tool_to_chatglm_function(tool)
            tools_json.append(tool_config)

        ans.append({
            'role': 'system',
            'content':
            'Answer the following questions as best as you can. You have access to the following tools:',
            'tools': tools_json
        })
        query = f"""{prompt.split("Human: ")[-1].strip()}"""
        return ans, query

    def _extract_observation(self, prompt: str):
        return_json = prompt.split('Observation: ')[-1].split('\nThought:')[0]
        self.history.append({'role': 'observation', 'content': return_json})
        return

    def _extract_tool(self):
        tool_names = list([t.name for t in self.tools])
        if self.history[-1]['metadata'] and len(self.history[-1]['metadata']) > 0:
            metadata = self.history[-1]['metadata']
            content = self.history[-1]['content']
            if 'tool_call' in content:
                for tool in tool_names:
                    if tool in metadata:
                        input_para = content.split("='")[-1].split("'")[0]
                        action_json = {'action': tool, 'action_input': input_para}
                        self.has_search = True
                        return f"""
Action:
```
{json.dumps(action_json, ensure_ascii=False)}
```"""

        final_answer_json = {'action': 'Final Answer', 'action_input': self.history[-1]['content']}
        self.has_search = False
        return f"""
Action:
```
{json.dumps(final_answer_json, ensure_ascii=False)}
```"""

    def get_full_inputs(self, intermediate_steps: List[Tuple[AgentAction, str]],
                        **kwargs: Any) -> Dict[str, Any]:
        """Create the full inputs for the LLMChain from intermediate steps."""
        thoughts = self._construct_scratchpad(intermediate_steps)
        new_inputs = {'agent_scratchpad': thoughts, 'stop': self._stop}
        full_inputs = {**kwargs, **new_inputs}
        return full_inputs

    def plan(
        self,
        intermediate_steps: List[Tuple[AgentAction, str]],
        callbacks: Callbacks = None,
        **kwargs: Any,
    ) -> Union[AgentAction, AgentFinish]:
        """Given input, decided what to do.

        Args:
            intermediate_steps: Steps the LLM has taken to date,
                along with observations
            callbacks: Callbacks to run.
            **kwargs: User inputs.

        Returns:
            Action specifying what tool to use.
        """
        full_inputs = self.get_full_inputs(intermediate_steps, **kwargs)
        print(f"----full_inputs{full_inputs}")
        prompt = self.prompt.format_prompt(**full_inputs)
        prompt = prompt.to_string()
        print(f"----prompt:{prompt}")
        if not self.has_search:
            self.history, query = self._tool_history(prompt)
        else:
            self._extract_observation(prompt)
            query = ''
        self.history.append({'role': 'user', 'content': query})
        print(f"----self.history:{self.history}")

        chat_messages = []
        for message in self.history:
            additional_kwargs = dict()
            for key in message.keys():
                if key not in ['role', 'content']:
                    additional_kwargs[key] = message[key]
            chat_messages.append(
                ChatMessage(role=message['role'],
                            content=message['content'],
                            additional_kwargs=additional_kwargs))

        predicted_message = self.llm.predict_messages(chat_messages, callbacks=callbacks)
        print(f"----predicted_message:{predicted_message}")
        self.history.append({
            'role': 'assistant',
            'content': predicted_message.content,
            'metadata': predicted_message.additional_kwargs['metadata']
        })
        full_output = self._extract_tool()
        print(f"----full_output:{full_output}")
        return self.output_parser.parse(full_output)

    async def aplan(
        self,
        intermediate_steps: List[Tuple[AgentAction, str]],
        callbacks: Callbacks = None,
        **kwargs: Any,
    ) -> Union[AgentAction, AgentFinish]:
        """Given input, decided what to do.

        Args:
            intermediate_steps: Steps the LLM has taken to date,
                along with observations
            callbacks: Callbacks to run.
            **kwargs: User inputs.

        Returns:
            Action specifying what tool to use.
        """
        full_inputs = self.get_full_inputs(intermediate_steps, **kwargs)
        prompt = self.prompt.format_prompt(**full_inputs)
        prompt = prompt.to_string()
        if not self.has_search:
            self.history, query = self._tool_history(prompt)
        else:
            self._extract_observation(prompt)
            query = ''
        self.history.append({'role': 'user', 'content': query})

        chat_messages = []
        for message in self.history:
            additional_kwargs = dict()
            for key in message.keys():
                if key not in ['role', 'content']:
                    additional_kwargs[key] = message[key]
            chat_messages.append(
                ChatMessage(role=message['role'],
                            content=message['content'],
                            additional_kwargs=additional_kwargs))

        predicted_message = await self.llm.apredict_messages(chat_messages, callbacks=callbacks)
        self.history.append({
            'role': 'assistant',
            'content': predicted_message.content,
            'metadata': predicted_message.additional_kwargs['metadata']
        })
        full_output = self._extract_tool()
        return self.output_parser.parse(full_output)

    @classmethod
    def create_prompt(
        cls,
        tools: Sequence[BaseTool],
        prefix: str = PREFIX,
        suffix: str = SUFFIX,
        human_message_template: str = HUMAN_MESSAGE_TEMPLATE,
        format_instructions: str = FORMAT_INSTRUCTIONS,
        input_variables: Optional[List[str]] = None,
        memory_prompts: Optional[List[BasePromptTemplate]] = None,
    ) -> BasePromptTemplate:
        tool_strings = []
        for tool in tools:
            args_schema = re.sub('}', '}}}}', re.sub('{', '{{{{', str(tool.args)))
            tool_strings.append(f'{tool.name}: {tool.description}, args: {args_schema}')
        formatted_tools = '\n'.join(tool_strings)
        tool_names = ', '.join([tool.name for tool in tools])
        format_instructions = format_instructions.format(tool_names=tool_names)
        template = '\n\n'.join([prefix, formatted_tools, format_instructions, suffix])
        if input_variables is None:
            input_variables = ['input', 'agent_scratchpad']
        _memory_prompts = memory_prompts or []
        messages = [
            SystemMessagePromptTemplate.from_template(template),
            *_memory_prompts,
            HumanMessagePromptTemplate.from_template(human_message_template),
        ]
        return ChatPromptTemplate(input_variables=input_variables, messages=messages)

    @classmethod
    def from_llm_and_tools(
        cls,
        llm: BaseLanguageModel,
        tools: Sequence[BaseTool],
        callback_manager: Optional[BaseCallbackManager] = None,
        output_parser: Optional[AgentOutputParser] = None,
        prefix: str = PREFIX,
        suffix: str = SUFFIX,
        human_message_template: str = HUMAN_MESSAGE_TEMPLATE,
        format_instructions: str = FORMAT_INSTRUCTIONS,
        input_variables: Optional[List[str]] = None,
        memory_prompts: Optional[List[BasePromptTemplate]] = None,
        **kwargs: Any,
    ) -> Agent:
        """Construct an agent from an LLM and tools."""
        cls._validate_tools(tools)
        prompt = cls.create_prompt(
            tools,
            prefix=prefix,
            suffix=suffix,
            human_message_template=human_message_template,
            format_instructions=format_instructions,
            input_variables=input_variables,
            memory_prompts=memory_prompts,
        )
        _output_parser = output_parser or cls._get_default_output_parser(llm=llm)
        return cls(
            llm=llm,
            prompt=prompt,
            tools=tools,
            output_parser=_output_parser,
            callback_manager=callback_manager,
            **kwargs,
        )

    @property
    def _agent_type(self) -> str:
        raise ValueError



import os
import pdb
import pandas as pd
import requests
import json
from config import API_KEY, DATASET_API, DIFY_TOKEN
from loguru import logger
import tqdm

datasets = set()
dataset = {}

def add_dataset(name, dataset_id):
    """    根据dataset 创建set， 映射name到id    """
    datasets.add(name)
    dataset[name] = dataset_id
    return dataset_id

def query_dataset_from_dify():
    """
    知识库列表
    curl --location --request GET 'https://api.dify.ai/v1/datasets?page=1&limit=20' \
    --header 'Authorization: Bearer {api_key}'"""
    url = 'https://api.dify.ai/v1/datasets?page=1&limit=100'
    headers = {
        'Authorization': f'Bearer {DATASET_API}', 
    }
    response = requests.get(url, headers=headers).json()
    datasets = response['data']
    print(datasets)
    
    for dataset in datasets:
        name = dataset['name']
        id = dataset['id']
        add_dataset(name, id)
    return datasets

def create_dataset(name):
    url = 'https://api.dify.ai/v1/datasets'
    headers = {
        'Authorization': f'Bearer {DATASET_API}',
        'Content-Type': 'application/json',
    }
    data = {
        'name': name,
    }
    response = requests.post(url, headers=headers, data=json.dumps(data)).json()
    print(response)
    return response

def query_file_from_dataset(dataset_id):
    """curl --location --request GET 'https://api.dify.ai/v1/datasets/{dataset_id}/documents' \
--header 'Authorization: Bearer {api_key}'"""
    url = f'https://api.dify.ai/v1/datasets/{dataset_id}/documents'
    headers = {
        'Authorization': f'Bearer {DATASET_API}',
    }
    response = requests.get(url, headers=headers).json()
    print(response)
    
    files = response['data']
    return files
    
def create_by_file(file_path, dataset_id):
    url = f'https://api.dify.ai/v1/datasets/{dataset_id}/document/create_by_file'
    headers = {
        'Authorization': f'Bearer {DATASET_API}',
    }
    data = {
        "name": "Dify",
        "indexing_technique": "high_quality",
        "process_rule": {
            "rules": {
                "pre_processing_rules": [
                    {"id": "remove_extra_spaces", "enabled": True},
                    {"id": "remove_urls_emails", "enabled": True}
                ],
                "segmentation": {
                    "separator": "###",
                    "max_tokens": 500
                }
            },
            "mode": "automatic"
        }
    }
    files = {
        'data': (None, json.dumps(data), 'application/json'),
        'file': open(file_path, 'rb')
    }
    response = requests.post(url, headers=headers, files=files).json()
    print(response)
    return response

def get_answer_from_dify(query, conversation_id="", user="abc123"):
    url = 'https://api.dify.ai/v1/chat-messages'
    headers = {
        'Authorization': f'Bearer {API_KEY}',
        'Content-Type': 'application/json',
    }
    data = {
        'inputs': {},
        'query': query,
        'response_mode': 'blocking',
        'conversation_id': conversation_id,
        'user': user,
        # 'files': [
        #     {
        #         "type": "image",
        #         "transfer_method": "local_file",
        #         "upload_file_id": f'{file_id}'}    
        # ],
    }
    response = requests.post(url, headers=headers, data=json.dumps(data)).json()
    print(response)
    return response

def upload_file_to_dify(data_dir, excel_file, save_excel_file):
    df = pd.read_excel(excel_file)
    all_questions_info = list()
    # 遍历每一行
    for index, row in df.iterrows():
        # 遍历每一列
        question_info = dict()
        for column in df.columns:
            value = row[column]
            question_info[column] = value
            # print(f"Row {index}, Column {column} has value {value}")
        all_questions_info.append(question_info)
    
    for questions_info in tqdm(all_questions_info):
        collection_name = questions_info['知识库名']
        file_name = questions_info['文件名']
        file_type = questions_info['文件类型']
        
        # pdb.set_trace()
        if file_type == 'pdf':
            # 没有知识库先创建
            if collection_name not in datasets:
                res = create_dataset(collection_name)
                dataset_id = res['id']
                add_dataset(collection_name, dataset_id) 
            else:
                dataset_id = dataset[collection_name]
           
            #  查询知识库中是否有该文件
            files = query_file_from_dataset(dataset_id)
            has_file = False
            for file in files:
                if file_name == file['name']:
                    has_file = True
                    file_id = file['id']
                    break
            if not has_file:
                try:
                    res = create_by_file(os.path.join(data_dir, file_name), dataset_id)
                    file_id = res['document']['id']
                except Exception as e:
                    file_id = ''
                    logger.warning(f'error in create file {file_name}: {e}')
                    
            questions_info['dataset_id'] = dataset_id           
        
    # save excel
    df = pd.DataFrame(all_questions_info)
    df.to_excel(save_excel_file, index=False)

def re_config_dataset(dataset_id, app_id='8ab350f2-d4f9-40c9-a9f3-ff5575332dea'):
    """抓包的应用config url"""
    url = f'https://cloud.dify.ai/console/api/apps/{app_id}/model-config'
    headers = {
        'authority': 'cloud.dify.ai',
        'accept': '*/*',
        'accept-language': 'zh-CN,zh;q=0.9',
        'authorization': f'Bearer {DIFY_TOKEN}',
        'content-type': 'application/json',
        'origin': 'https://cloud.dify.ai',
        'referer': 'https://cloud.dify.ai/app/97b613df-c950-4c53-a9ae-125bd1c54386/configuration',
        'sec-ch-ua': '"Not_A Brand";v="8", "Chromium";v="120", "Google Chrome";v="120"',
        'sec-ch-ua-mobile': '?0',
        'sec-ch-ua-platform': '"Windows"',
        'sec-fetch-dest': 'empty',
        'sec-fetch-mode': 'cors',
        'sec-fetch-site': 'same-origin',
        'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36',
    }
    data = {
        "pre_prompt": None,
        "prompt_type": "simple",
        "chat_prompt_config": {},
        "completion_prompt_config": {},
        "user_input_form": [],
        "dataset_query_variable": "",
        "opening_statement": "",
        "more_like_this": {"enabled": False},
        "suggested_questions_after_answer": {"enabled": False},
        "speech_to_text": {"enabled": False},
        "retriever_resource": {"enabled": False},
        "sensitive_word_avoidance": {"enabled": False, "type": "", "configs": []},
        "external_data_tools": [],
        "agent_mode": {
            "enabled": True,
            "tools": [
                {"dataset": {"enabled": True, "id": f'{dataset_id}'}},
            ]
        },
        "model": {
            "provider": "openai",
            "name": "gpt-4-1106-preview",
            "mode": "chat",
            "completion_params": {"max_tokens": 512, "temperature": 1, "top_p": 1, "presence_penalty": 0, "frequency_penalty": 0}
        },
        "dataset_configs": {"retrieval_model": "single", "top_k": 20, "score_threshold": 0.5},
        "file_upload": {"image": {"enabled": False, "number_limits": 3, "detail": "high", "transfer_methods": ["remote_url", "local_file"]}}
    }

    response = requests.post(url, headers=headers, data=json.dumps(data))
    print(response.json())
               
    
def question_answer(excel_file):
    df = pd.read_excel(excel_file)
    all_questions_info = list()
    # 遍历每一行
    for index, row in df.iterrows():
        # 遍历每一列
        question_info = dict()
        for column in df.columns:
            value = row[column]
            question_info[column] = value
            # print(f"Row {index}, Column {column} has value {value}")
        all_questions_info.append(question_info)  

    for questions_info in all_questions_info:
        question = questions_info['问题']
        file_type = questions_info['文件类型']
        collection_name = questions_info['知识库名']
        dataset_id = questions_info['dataset_id']
        if collection_name not in datasets or dataset_id == '':
            res = create_dataset(collection_name)
            dataset_id = add_dataset(collection_name, res['id']) 
        
        try:
            re_config_dataset(dataset_id)
            response = get_answer_from_dify(question)
            ans = response['answer']
        except Exception as e:
            print(f'error in query {question}')
            ans = ''    
        print('ans:', ans, 'question:', question)
        questions_info['dify_answer'] = ans
    
    df = pd.DataFrame(all_questions_info)
    df.to_excel(os.path.join('questions_info_with_answer_dify.xlsx'), index=False)

if __name__ == '__main__':
    data_dir = './rag_benchmark_v1.0/rag_benchmark'
    save_dir = './rag_benchmark_v1.0/rag_benchmark_processed'
    excel_file = './data/questions_info_with_answer_sample.xlsx'
    save_excel_file = './data/questions_info_with_dify_file_id.xlsx'
    
    query_dataset_from_dify()
    # upload_file_to_dify(save_dir, excel_file, save_excel_file)
    question_answer(save_excel_file)
    



from langchain.prompts.prompt import PromptTemplate


EXTRACT_KEY_PROMPT = PromptTemplate(
    input_variables=['question'],
    template="""分析给定Question，提取Question中包含的KeyWords，输出列表形式

Examples:
Question: 达梦公司在过去三年中的流动比率如下：2021年：3.74倍；2020年：2.82倍；2019年：2.05倍。
KeyWords: ['过去三年', '流动比率', '2021', '3.74', '2020', '2.82', '2019', '2.05']

----------------
Question: {question}
KeyWords: """,
)

# EXTRACT_KEY_PROMPT = PromptTemplate(
#     input_variables=['question'],
#     template="""分析给定Question，提取Question中包含的KeyWords，输出列表形式

# Examples:
# Question: 能否根据2020年金宇生物技术股份有限公司的年报，给我简要介绍一下报告期内公司的社会责任工作情况？
# KeyWords: ['报告期', '社会责任', '工作情况']

# Question: 请根据江化微2019年的年报，简要介绍报告期内公司主要销售客户的客户集中度情况，并结合同行业情况进行分析。
# KeyWords: ['报告期', '主要', '销售客户', '客户集中度', '同行业', '分析']

# Question: 请问，在苏州迈为科技股份有限公司2019年的年报中，现金流的情况是否发生了重大变化？若发生，导致重大变化的原因是什么？
# KeyWords: ['现金流', '重大变化', '原因']

# ----------------
# Question: {question}
# KeyWords: """,
# )


"""tianyancha api"""
from __future__ import annotations

from typing import Any, Dict, Type

from bisheng_langchain.utils.requests import Requests, RequestsWrapper
from langchain_core.pydantic_v1 import BaseModel, Field, root_validator

from .base import APIToolBase


class InputArgs(BaseModel):
    """args_schema"""
    query: str = Field(description='搜索关键字（公司名称、公司id、注册号或社会统一信用代码）')


class CompanyInfo(APIToolBase):
    """Manage tianyancha company client."""
    api_key: str = None
    args_schema: Type[BaseModel] = InputArgs

    @root_validator(pre=True)
    def build_header(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Build headers that were passed in."""
        if not values.get('api_key'):
            raise ValueError('Parameters api_key should be specified give.')

        headers = values.get('headers', {})
        headers.update({'Authorization': values['api_key']})
        values['headers'] = headers
        return values

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key and python package exists in environment."""
        timeout = values.get('request_timeout', 30)
        if not values.get('client'):
            values['client'] = Requests(headers=values['headers'], request_timeout=timeout)
        if not values.get('async_client'):
            values['async_client'] = RequestsWrapper(headers=values['headers'],
                                                     request_timeout=timeout)
        return values

    @classmethod
    def search_company(cls, api_key: str, pageNum: int = 1, pageSize: int = 20) -> CompanyInfo:
        """可以通过关键词获取企业列表，企业列表包括公司名称或ID、类型、成立日期、经营状态、统一社会信用代码等字段的详细信息"""
        url = 'http://open.api.tianyancha.com/services/open/search/2.0'
        input_key = 'word'
        params = {}
        params['pageSize'] = pageSize
        params['pageNum'] = pageNum

        return cls(url=url, api_key=api_key, input_key=input_key, params=params)

    @classmethod
    def get_company_baseinfo(cls, api_key: str) -> CompanyInfo:
        """可以通过公司名称或ID获取企业基本信息，企业基本信息包括公司名称或ID、类型、成立日期、经营状态、注册资本、法人、工商注册号、统一社会信用代码、组织机构代码、纳税人识别号等字段信息"""
        url = 'http://open.api.tianyancha.com/services/open/ic/baseinfo/normal'
        input_key = 'keyword'
        params = {}

        return cls(url=url, api_key=api_key, input_key=input_key, params=params)

    @classmethod
    def ip_rights(cls, api_key: str) -> CompanyInfo:
        """可以通过公司名称或ID获取包含商标、专利、作品著作权、软件著作权、网站备案等维度的相关信息"""
        url = 'http://open.api.tianyancha.com/services/open/cb/ipr/3.0'
        input_key = 'keyword'

        return cls(url=url, api_key=api_key, input_key=input_key, params={})

    @classmethod
    def judicial_risk(cls, api_key: str) -> CompanyInfo:
        """可以通过公司名称或ID获取包含法律诉讼、法院公告、开庭公告、失信人、被执行人、立案信息、送达公告等维度的相关信息"""
        url = 'http://open.api.tianyancha.com/services/open/cb/judicial/2.0'
        return cls(url=url, api_key=api_key)

    @classmethod
    def ic_info(cls, api_key: str) -> CompanyInfo:
        """可以通过公司名称或ID获取包含企业基本信息、主要人员、股东信息、对外投资、分支机构等维度的相关信息"""
        url = 'http://open.api.tianyancha.com/services/open/cb/ic/2.0'

        return cls(url=url, api_key=api_key)

    @classmethod
    def law_suit_case(cls, api_key: str, pageSize: int = 20, pageNum: int = 1) -> CompanyInfo:
        """可以通过公司名称或ID获取企业法律诉讼信息，法律诉讼包括案件名称、案由、案件身份、案号等字段的详细信息"""
        url = 'http://open.api.tianyancha.com/services/open/jr/lawSuit/3.0'
        params = {}
        params['pageSize'] = pageSize
        params['pageNum'] = pageNum
        return cls(url=url, api_key=api_key, params=params)

    @classmethod
    def company_change_info(cls,
                            api_key: str,
                            pageSize: int = 20,
                            pageNum: int = 1) -> CompanyInfo:
        """可以通过公司名称或ID获取企业变更记录，变更记录包括工商变更事项、变更前后信息等字段的详细信息"""
        url = 'http://open.api.tianyancha.com/services/open/ic/changeinfo/2.0'
        params = {}
        params['pageSize'] = pageSize
        params['pageNum'] = pageNum
        return cls(url=url, api_key=api_key, params=params)

    @classmethod
    def company_holders(cls, api_key: str, pageSize: int = 20, pageNum: int = 1) -> CompanyInfo:
        """可以通过公司名称或ID获取企业股东信息，股东信息包括股东名、出资比例、出资金额、股东总数等字段的详细信息"""
        url = 'http://open.api.tianyancha.com/services/open/ic/holder/2.0'
        params = {}
        params['pageSize'] = pageSize
        params['pageNum'] = pageNum
        return cls(url=url, api_key=api_key, params=params)

    @classmethod
    def all_companys_by_company(cls, api_key: str, pageSize: int = 20, pageNum: int = 1):
        """可以通过公司名称和人名获取企业人员的所有相关公司，包括其担任法人、股东、董监高的公司信息"""
        url = 'http://open.api.tianyancha.com/services/v4/open/allCompanys'
        input_key = 'humanName'
        params = {}
        params['pageSize'] = pageSize
        params['pageNum'] = pageNum

        class InputArgs(BaseModel):
            """args_schema"""
            query: str = Field(description='human who you want to search')
            name: str = Field(description='company name which human worked')

        return cls(url=url,
                   api_key=api_key,
                   params=params,
                   input_key=input_key,
                   args_schema=InputArgs)

    @classmethod
    def riskinfo(cls, api_key: str) -> CompanyInfo:
        """可以通过关键字（公司名称、公司id、注册号或社会统一信用代码）获取企业相关天眼风险列表，包括企业自身/周边/预警风险信息。"""
        url = 'http://open.api.tianyancha.com/services/open/risk/riskInfo/2.0'
        return cls(url=url, api_key=api_key)



from typing import List  # 导入 List 类型注解
from langchain.callbacks.manager import (AsyncCallbackManagerForRetrieverRun, CallbackManagerForRetrieverRun)  # 导入 Langchain 回调管理器
from langchain.schema import BaseRetriever, Document  # 导入 Langchain 基础检索器和文档模型

class JiShiJiaoEsVectorRetriever(BaseRetriever):
    """
    JiShiJiaoEsVectorRetriever 类，用于组合 Elasticsearch 关键词检索器和向量检索器的结果。
    此类集成了 Elasticsearch 关键词检索和向量检索的结果，通过不同的组合策略，
    可以灵活地融合两种检索方式的优点，提升检索效果。
    Args:
        vector_retriever (BaseRetriever): 向量检索器实例，用于执行向量相似度搜索。
        keyword_retriever (BaseRetriever): 关键词检索器实例，用于执行关键词匹配搜索。
        combine_strategy (str): 组合策略，用于指定如何组合关键词检索和向量检索的结果，可选值包括：
            - 'keyword_front': 关键词检索结果在前，向量检索结果在后。
            - 'vector_front': 向量检索结果在前，关键词检索结果在后。
            - 'mix': 混合组合，交替排列关键词检索和向量检索的结果。
            默认为 'keyword_front'。
    """
    vector_retriever: BaseRetriever  # 向量检索器
    keyword_retriever: BaseRetriever  # 关键词检索器
    combine_strategy: str = 'keyword_front'  # 组合策略，默认为 'keyword_front'，关键词检索结果在前

    def _get_relevant_documents(self, query: str, *, run_manager: CallbackManagerForRetrieverRun, ) -> List[Document]:
        """
        获取给定查询的相关文档列表 (同步方法).
        根据指定的组合策略，融合关键词检索器和向量检索器的检索结果，返回最终的相关文档列表。
        Args:
            query (str): 查询语句。
            run_manager (CallbackManagerForRetrieverRun): 回调管理器，用于处理检索过程中的回调事件。
        Returns:
            List[Document]: 相关文档列表。
        """
        print("jishijiao - _get_relevant_documents")  # 打印日志，标记进入同步检索方法
        # 获取检索器的融合结果。
        vector_docs = self.vector_retriever.get_relevant_documents(query, callbacks=run_manager.get_child())  # 调用向量检索器获取相关文档
        keyword_docs = self.keyword_retriever.get_relevant_documents(query, callbacks=run_manager.get_child())  # 调用关键词检索器获取相关文档
        if self.combine_strategy == 'keyword_front':  # 如果组合策略为 'keyword_front'
            return keyword_docs + vector_docs  # 关键词检索结果在前，向量检索结果在后
        elif self.combine_strategy == 'vector_front':  # 如果组合策略为 'vector_front'
            return vector_docs + keyword_docs  # 向量检索结果在前，关键词检索结果在后
        elif self.combine_strategy == 'mix':  # 如果组合策略为 'mix'
            combine_docs_dict = {}  # 初始化合并文档字典，用于去重
            min_len = min(len(keyword_docs), len(vector_docs))  # 获取关键词和向量检索结果列表的最小长度
            for i in range(min_len):  # 遍历较短的结果列表长度
                combine_docs_dict[keyword_docs[i].page_content] = keyword_docs[i]  # 将关键词检索结果添加到合并文档字典
                combine_docs_dict[vector_docs[i].page_content] = vector_docs[i]  # 将向量检索结果添加到合并文档字典
            for doc in keyword_docs[min_len:]:  # 将关键词检索结果列表中剩余的文档添加到合并文档字典
                combine_docs_dict[doc.page_content] = doc
            for doc in vector_docs[min_len:]:  # 将向量检索结果列表中剩余的文档添加到合并文档字典
                combine_docs_dict[doc.page_content] = doc
            # 将字典的值转换为列表，实现混合组合，并去重
            combine_docs = list(combine_docs_dict.values())  # 将合并文档字典的值转换为列表
            return combine_docs  # 返回合并后的文档列表
        else:  # 如果组合策略不是预定义的值
            raise ValueError(f'Expected combine_strategy to be one of '
                             f'(keyword_front, vector_front, mix),'
                             f'instead found {self.combine_strategy}')  # 抛出 ValueError 异常，提示组合策略无效

    async def _aget_relevant_documents(self, query: str, *, run_manager: AsyncCallbackManagerForRetrieverRun, ) -> List[Document]:
        """
        异步获取给定查询的相关文档列表 (异步方法).
        根据指定的组合策略，异步融合关键词检索器和向量检索器的检索结果，返回最终的相关文档列表。
        Args:
            query (str): 查询语句。
            run_manager (AsyncCallbackManagerForRetrieverRun): 异步回调管理器，用于处理检索过程中的回调事件。
        Returns:
            List[Document]: 相关文档列表。
        """
        print("jishijiao - _aget_relevant_documents")  # 打印日志，标记进入异步检索方法
        # 获取检索器的融合结果。
        vector_docs = await self.vector_retriever.aget_relevant_documents(query, callbacks=run_manager.get_child())  # 异步调用向量检索器获取相关文档
        keyword_docs = await self.keyword_retriever.aget_relevant_documents(query, callbacks=run_manager.get_child())  # 异步调用关键词检索器获取相关文档
        if self.combine_strategy == 'keyword_front':  # 如果组合策略为 'keyword_front'
            return keyword_docs + vector_docs  # 关键词检索结果在前，向量检索结果在后
        elif self.combine_strategy == 'vector_front':  # 如果组合策略为 'vector_front'
            return vector_docs + keyword_docs  # 向量检索结果在前，关键词检索结果在后
        elif self.combine_strategy == 'mix':  # 如果组合策略为 'mix'
            combine_docs_dict = {}  # 初始化合并文档字典，用于去重
            min_len = min(len(keyword_docs), len(vector_docs))  # 获取关键词和向量检索结果列表的最小长度
            for i in range(min_len):  # 遍历较短的结果列表长度
                combine_docs_dict[keyword_docs[i].page_content] = keyword_docs[i]  # 将关键词检索结果添加到合并文档字典
                combine_docs_dict[vector_docs[i].page_content] = vector_docs[i]  # 将向量检索结果添加到合并文档字典
            for doc in keyword_docs[min_len:]:  # 将关键词检索结果列表中剩余的文档添加到合并文档字典
                combine_docs_dict[doc.page_content] = doc
            for doc in vector_docs[min_len:]:  # 将向量检索结果列表中剩余的文档添加到合并文档字典
                combine_docs_dict[doc.page_content] = doc
            # 将字典的值转换为列表，实现混合组合，并去重
            combine_docs = list(combine_docs_dict.values())  # 将合并文档字典的值转换为列表
            return combine_docs  # 返回合并后的文档列表
        else:  # 如果组合策略不是预定义的值
            raise ValueError(f'Expected combine_strategy to be one of '
                             f'(keyword_front, vector_front, mix),'
                             f'instead found {self.combine_strategy}')  # 抛出 ValueError 异常，提示组合策略无效



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Collection, Dict, List
from venv import logger

import requests
from langchain.schema.document import Document
from langchain.vectorstores.base import VectorStore, VectorStoreRetriever
from langchain_core.pydantic_v1 import Field, root_validator

if TYPE_CHECKING:
    from langchain.callbacks.manager import (
        AsyncCallbackManagerForRetrieverRun,
        CallbackManagerForRetrieverRun,
    )


class VectorStoreFilterRetriever(VectorStoreRetriever):
    vectorstore: VectorStore
    search_type: str = 'similarity'
    search_kwargs: dict = Field(default_factory=dict)
    allowed_search_types: ClassVar[Collection[str]] = (
        'similarity',
        'similarity_score_threshold',
        'mmr',
    )
    access_url: str = None

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    @root_validator()
    def validate_search_type(cls, values: Dict) -> Dict:
        """Validate search type."""
        search_type = values['search_type']
        if search_type not in cls.allowed_search_types:
            raise ValueError(f'search_type of {search_type} not allowed. Valid values are: '
                             f'{cls.allowed_search_types}')
        if search_type == 'similarity_score_threshold':
            score_threshold = values['search_kwargs'].get('score_threshold')
            if (score_threshold is None) or (not isinstance(score_threshold, float)):
                raise ValueError('`score_threshold` is not specified with a float value(0~1) '
                                 'in `search_kwargs`.')
        return values

    def _get_relevant_documents(self, query: str, *,
                                run_manager: CallbackManagerForRetrieverRun) -> List[Document]:
        if self.search_type == 'similarity':
            docs = self.vectorstore.similarity_search(query, **self.search_kwargs)
        elif self.search_type == 'similarity_score_threshold':
            docs_and_similarities = (self.vectorstore.similarity_search_with_relevance_scores(
                query, **self.search_kwargs))
            docs = [doc for doc, _ in docs_and_similarities]
        elif self.search_type == 'mmr':
            docs = self.vectorstore.max_marginal_relevance_search(query, **self.search_kwargs)
        else:
            raise ValueError(f'search_type of {self.search_type} not allowed.')

        return self.get_file_access(docs)

    async def _aget_relevant_documents(
            self, query: str, *,
            run_manager: AsyncCallbackManagerForRetrieverRun) -> List[Document]:
        if self.search_type == 'similarity':
            docs = await self.vectorstore.asimilarity_search(query, **self.search_kwargs)
        elif self.search_type == 'similarity_score_threshold':
            docs_and_similarities = (await
                                     self.vectorstore.asimilarity_search_with_relevance_scores(
                                         query, **self.search_kwargs))
            docs = [doc for doc, _ in docs_and_similarities]
        elif self.search_type == 'mmr':
            docs = await self.vectorstore.amax_marginal_relevance_search(
                query, **self.search_kwargs)
        else:
            raise ValueError(f'search_type of {self.search_type} not allowed.')
        return self.get_file_access(docs)

    def get_file_access(self, docs: List[Document]):
        file_ids = [doc.metadata.get('file_id') for doc in docs if 'file_id' in doc.metadata]
        if file_ids:
            res = requests.get(self.access_url, json=file_ids)
            if res.status_code == 200:
                doc_res = res.json().get('data') or []
                doc_right = {doc.get('docid') for doc in doc_res if doc.get('result') == 1}
                for doc in docs:
                    if doc.metadata.get('file_id') and doc.metadata.get(
                            'file_id') not in doc_right:
                        doc.page_content = ''
                        doc.metadata['right'] = False
                return docs
            else:
                logger.error(f'query_file_access_fail url={self.access_url} res={res.text}')
                return [Document(page_content='', metadata={})]
        else:
            return docs



from typing import Any, List, Tuple

from langchain.callbacks.manager import Callbacks
from langchain.chains.combine_documents.stuff import StuffDocumentsChain as StuffDocumentsChainOld
from langchain.docstore.document import Document


class StuffDocumentsChain(StuffDocumentsChainOld):

    token_max: int = -1

    def combine_docs(self,
                     docs: List[Document],
                     callbacks: Callbacks = None,
                     **kwargs: Any) -> Tuple[str, dict]:
        """Stuff all documents into one prompt and pass to LLM.

        Args:
            docs: List of documents to join together into one variable
            callbacks: Optional callbacks to pass along
            **kwargs: additional parameters to use to get inputs to LLMChain.

        Returns:
            The first element returned is the single string output. The second
            element returned is a dictionary of other keys to return.
        """
        inputs = self._get_inputs(docs, **kwargs)
        # print('inputs:', len(inputs['context']))
        # print('prompt_length:', self.prompt_length(docs, **kwargs))
        if self.token_max > 0:
            inputs[self.document_variable_name] = inputs[
                self.document_variable_name][:self.token_max]
        # Call predict on the LLM.
        return self.llm_chain.predict(callbacks=callbacks, **inputs), {}

    async def acombine_docs(self,
                            docs: List[Document],
                            callbacks: Callbacks = None,
                            **kwargs: Any) -> Tuple[str, dict]:
        """Stuff all documents into one prompt and pass to LLM.

        Args:
            docs: List of documents to join together into one variable
            callbacks: Optional callbacks to pass along
            **kwargs: additional parameters to use to get inputs to LLMChain.

        Returns:
            The first element returned is the single string output. The second
            element returned is a dictionary of other keys to return.
        """
        inputs = self._get_inputs(docs, **kwargs)
        if self.token_max > 0:
            inputs[self.document_variable_name] = inputs[
                self.document_variable_name][:self.token_max]
        # Call predict on the LLM.
        return await self.llm_chain.apredict(callbacks=callbacks, **inputs), {}



import base64
import json
# import time
from typing import List, Optional

import requests
from langchain.document_loaders.blob_loaders import Blob
from langchain.schema import Document


class LayoutParser(object):
    """Parse image layout structure.
    """

    def __init__(self,
                 api_key: Optional[str] = None,
                 api_base_url: Optional[str] = None):
        self.api_key = api_key
        self.api_base_url = 'http://192.168.106.20:14569/predict'
        self.class_name = ['印章', '图片', '标题', '段落', '表格', '页眉', '页码', '页脚']

    def parse(self, blob: Blob) -> List[Document]:
        b64_data = base64.b64encode(blob.as_bytes()).decode()
        data = {'img': b64_data}
        resp = requests.post('http://192.168.106.20:14569/predict', data=data)
        content = resp.json()
        doc = Document(page_content=json.dumps(content), metadata={})
        return [doc]



from bisheng_langchain.gpts.agent_types.llm_functions_agent import (
    get_openai_functions_agent_executor, 
    get_qwen_local_functions_agent_executor
)
from bisheng_langchain.gpts.agent_types.llm_react_agent import get_react_agent_executor


__all__ = [
    "get_openai_functions_agent_executor",
    "get_qwen_local_functions_agent_executor",
    "get_react_agent_executor"
]


from typing import Any

from langchain_core.tools import BaseTool
from loguru import logger
from pydantic import BaseModel, create_model

from .base import APIToolBase, MultArgsSchemaTool, Field


class OpenApiTools(APIToolBase):

    def get_real_path(self):
        return self.url + self.params["path"]

    def get_request_method(self):
        return self.params["method"].lower()

    def get_params_json(self, **kwargs):
        params_define = {}
        for one in self.params["parameters"]:
            params_define[one["name"]] = one

        params = {}
        json_data = {}
        for k, v in kwargs.items():
            if params_define.get(k):
                if params_define[k]["in"] == "query":
                    params[k] = v
                else:
                    json_data[k] = v
            else:
                params[k] = v
        return params, json_data

    def parse_args_schema(self):
        params = self.params["parameters"]
        model_params = {}
        for one in params:
            field_type = one["schema"]["type"]
            if field_type == "number":
                field_type = "float"
            elif field_type == "integer":
                field_type = "int"
            elif field_type == "string":
                field_type = "str"
            else:
                raise Exception(f"schema type is not support: {field_type}")
            model_params[one["name"]] = (eval(field_type), Field(description=one["description"]))
        return create_model("InputArgs", __module__='bisheng_langchain.gpts.tools.api_tools.openapi',
                            __base__=BaseModel, **model_params)

    def run(self, **kwargs) -> str:
        """Run query through api and parse result."""
        path = self.get_real_path()
        logger.info('api_call url={}', path)
        method = self.get_request_method()
        params, json_data = self.get_params_json(**kwargs)

        if method == "get":
            resp = self.client.get(path, params=params)
        elif method == 'post':
            resp = self.client.post(path, params=params, json=json_data)
        elif method == 'put':
            resp = self.client.put(path, params=params, json=json_data)
        elif method == 'delete':
            resp = self.client.delete(path, params=params, json=json_data)
        else:
            raise Exception(f"http method is not support: {method}")
        if resp.status_code != 200:
            logger.info(f'api_call_fail code={resp.status_code} res={resp.text}')
            raise Exception(f"api_call_fail: {resp.status_code} {resp.text}")
        return resp.text

    async def arun(self, **kwargs) -> str:
        """Run query through api and parse result."""
        path = self.get_real_path()
        logger.info('api_call url={}', path)
        method = self.get_request_method()
        params, json_data = self.get_params_json(**kwargs)

        if method == "get":
            resp = await self.async_client.aget(path, params=params)
        elif method == 'post':
            resp = await self.async_client.apost(path, params=params, json=json_data)
        elif method == 'put':
            resp = await self.async_client.aput(path, params=params, json=json_data)
        elif method == 'delete':
            resp = await self.async_client.adelete(path, params=params, json=json_data)
        else:
            raise Exception(f"http method is not support: {method}")
        return resp

    @classmethod
    def get_api_tool(cls, name, **kwargs: Any) -> BaseTool:
        description = kwargs.pop("description", "")
        obj = cls(**kwargs)
        return MultArgsSchemaTool(name=name,
                                  description=description,
                                  func=obj.run,
                                  coroutine=obj.arun,
                                  args_schema=obj.parse_args_schema())



import asyncio
import logging
from enum import Enum
from functools import lru_cache
from typing import Any, Mapping, Optional, Sequence
from urllib.parse import urlparse

import httpx
import yaml
from bisheng_langchain.gpts.load_tools import get_all_tool_names, load_tools
from bisheng_langchain.gpts.utils import import_by_type, import_class
from langchain.tools import BaseTool
from langchain_core.language_models.base import LanguageModelLike
from langchain_core.messages import HumanMessage, AIMessage
from langchain_core.runnables import RunnableBinding

logger = logging.getLogger(__name__)


class ConfigurableAssistant(RunnableBinding):
    agent_executor_type: str
    tools: Sequence[BaseTool]
    llm: LanguageModelLike
    assistant_message: str
    interrupt_before_action: bool = False
    recursion_limit: int = 50

    def __init__(
        self,
        *,
        agent_executor_type: str,
        tools: Sequence[BaseTool],
        llm: LanguageModelLike,
        assistant_message: str,
        interrupt_before_action: bool = False,
        recursion_limit: int = 50,
        kwargs: Optional[Mapping[str, Any]] = None,
        config: Optional[Mapping[str, Any]] = None,
        **others: Any,
    ) -> None:
        others.pop("bound", None)
        agent_executor_object = import_class(f'bisheng_langchain.gpts.agent_types.{agent_executor_type}')

        _agent_executor = agent_executor_object(tools, llm, assistant_message, interrupt_before_action)
        agent_executor = _agent_executor.with_config({"recursion_limit": recursion_limit})
        super().__init__(
            agent_executor_type=agent_executor_type,
            tools=tools,
            llm=llm,
            assistant_message=assistant_message,
            bound=agent_executor,
            kwargs=kwargs or {},
            config=config or {},
        )


class BishengAssistant:

    def __init__(self, yaml_path) -> None:
        self.yaml_path = yaml_path
        with open(self.yaml_path, 'r') as f:
            self.params = yaml.safe_load(f)

        self.assistant_params = self.params['assistant']

        # init assistant prompt
        prompt_type = self.assistant_params['prompt_type']
        assistant_message = import_class(f'bisheng_langchain.gpts.prompts.{prompt_type}')

        # init llm or agent
        llm_params = self.assistant_params['llm']
        llm_object = import_by_type(_type='llms', name=llm_params['type'])
        if llm_params['type'] == 'ChatOpenAI' and llm_params['openai_proxy']:
            llm_params.pop('type')
            llm = llm_object(
                http_client=httpx.Client(proxies=llm_params['openai_proxy']),
                http_async_client=httpx.AsyncClient(proxies=llm_params['openai_proxy']),
                **llm_params,
            )
        else:
            llm_params.pop('type')
            llm = llm_object(**llm_params)

        # init tools
        available_tools = get_all_tool_names()
        tools = []
        for tool in self.assistant_params['tools']:
            tool_type = tool.pop('type')
            tool_config = tool if tool else {}
            if tool_type not in available_tools:
                raise ValueError(f"Tool type {tool_type} not found in TOOLS")
            _returned_tools = load_tools({tool_type: tool_config})
            if isinstance(_returned_tools, list):
                tools.extend(_returned_tools)
            else:
                tools.append(_returned_tools)

        # init agent executor
        agent_executor_params = self.assistant_params['agent_executor']
        self.agent_executor_type = agent_executor_params.pop('type')
        self.assistant = ConfigurableAssistant(
            agent_executor_type=self.agent_executor_type, 
            tools=tools, 
            llm=llm, 
            assistant_message=assistant_message, 
            **agent_executor_params
        )

    def run(self, query, chat_history=[], chat_round=5):
        if len(chat_history) % 2 != 0:
            raise ValueError("chat history should be even")
        
        # 限制chat_history轮数
        if len(chat_history) > chat_round * 2:
            chat_history = chat_history[-chat_round*2:]

        inputs = []
        for i in range(0, len(chat_history), 2):
            inputs.append(HumanMessage(content=chat_history[i]))
            inputs.append(AIMessage(content=chat_history[i+1]))
        inputs.append(HumanMessage(content=query))
        if self.agent_executor_type == 'get_react_agent_executor':
            result = asyncio.run(self.assistant.ainvoke({"input": inputs[-1].content, "chat_history": inputs[:-1]}))
        else:
            result = asyncio.run(self.assistant.ainvoke(inputs))
        return result


if __name__ == "__main__":
    from langchain.globals import set_debug

    # set_debug(True)
    # chat_history = []
    # query = "分析当日市场行情"
    chat_history = ['你好', '你好，有什么可以帮助你吗？', '福蓉科技股价多少?', '福蓉科技（股票代码：300049）的当前股价为48.67元。']
    query = '今天是什么时候？去年这个时候的股价是多少？'
    bisheng_assistant = BishengAssistant("config/base_scene.yaml")
    # bisheng_assistant = BishengAssistant("config/knowledge_scene.yaml")
    # bisheng_assistant = BishengAssistant("config/rag_scene.yaml")
    result = bisheng_assistant.run(query, chat_history=chat_history)
    print(result)



from typing import Any, Dict, Tuple, Type, Union

from bisheng_langchain.utils.requests import Requests, RequestsWrapper
from langchain_core.pydantic_v1 import BaseModel, Extra, Field, root_validator
from langchain_core.tools import BaseTool, Tool
from loguru import logger


class ApiArg(BaseModel):
    query: str = Field(description='query to look up in this tool')


class MultArgsSchemaTool(Tool):

    def _to_args_and_kwargs(self, tool_input: Union[str, Dict]) -> Tuple[Tuple, Dict]:
        # For backwards compatibility, if run_input is a string,
        # pass as a positional argument.
        if isinstance(tool_input, str):
            return (tool_input, ), {}
        else:
            return (), tool_input


class APIToolBase(BaseModel):
    """Manage tianyancha company client."""

    client: Any = Field(default=None, exclude=True)  #: :meta private:
    async_client: Any = Field(default=None, exclude=True)  #: :meta private:
    headers: Dict[str, Any] = {}
    request_timeout: int = 30
    url: str = None
    params: Dict[str, Any] = Field(default_factory=dict)
    input_key: str = 'keyword'
    args_schema: Type[BaseModel] = ApiArg

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key and python package exists in environment."""
        timeout = values.get('request_timeout', 30)
        if not values.get('client'):
            values['client'] = Requests(headers=values['headers'], request_timeout=timeout)
        if not values.get('async_client'):
            values['async_client'] = RequestsWrapper(headers=values['headers'],
                                                     request_timeout=timeout)
        return values

    def run(self, query: str, **kwargs) -> str:
        """Run query through api and parse result."""
        if query:
            self.params[self.input_key] = query
        if kwargs:
            self.params.update(kwargs)
        if self.params:
            param = '&'.join([f'{k}={v}' for k, v in self.params.items()])
            url = self.url + '?' + param if '?' not in self.url else self.url + '&' + param
        else:
            url = self.url
        logger.info('api_call url={}', url)
        resp = self.client.get(url)
        if resp.status_code != 200:
            logger.info('api_call_fail res={}', resp.text)
        return resp.text[:10000]

    async def arun(self, query: str, **kwargs) -> str:
        """Run query through api and parse result."""
        if query:
            self.params[self.input_key] = query
        if kwargs:
            self.params.update(kwargs)
        if self.params:
            param = '&'.join([f'{k}={v}' for k, v in self.params.items()])
            url = self.url + '?' + param if '?' not in self.url else self.url + '&' + param
        else:
            url = self.url
        logger.info('api_call url={}', url)
        resp = await self.async_client.aget(url)
        logger.info(resp[:10000])
        return resp[:10000]

    @classmethod
    def get_api_tool(cls, name, **kwargs: Any) -> BaseTool:
        attr_name = name.split('_', 1)[-1]
        class_method = getattr(cls, attr_name)

        return MultArgsSchemaTool(name=name,
                                  description=class_method.__doc__,
                                  func=class_method(**kwargs).run,
                                  coroutine=class_method(**kwargs).arun,
                                  args_schema=class_method(**kwargs).args_schema)



from bisheng_langchain.chains.autogen.auto_gen import AutoGenChain
from bisheng_langchain.chains.combine_documents.stuff import StuffDocumentsChain
from bisheng_langchain.chains.conversational_retrieval.base import ConversationalRetrievalChain
from bisheng_langchain.chains.retrieval.retrieval_chain import RetrievalChain
from bisheng_langchain.chains.router.multi_rule import MultiRuleChain
from bisheng_langchain.chains.router.rule_router import RuleBasedRouter
from bisheng_langchain.chains.transform import TransformChain
from bisheng_langchain.chains.qa_generation.base import QAGenerationChain
from bisheng_langchain.chains.qa_generation.base_v2 import QAGenerationChainV2

from .loader_output import LoaderOutputChain

__all__ = [
    'StuffDocumentsChain', 'LoaderOutputChain', 'AutoGenChain', 'RuleBasedRouter',
    'MultiRuleChain', 'RetrievalChain', 'ConversationalRetrievalChain', 'TransformChain',
    'QAGenerationChain', 'QAGenerationChainV2'
]
